"""Common util functions used by other packages in tealer.

Modules:
    code_complexity: Contains util functions to compute code complexity
        of the contracts.

    command_line: Defines output util functions used by command line
        argument handlers.

    comparable_enum: Defines a comparable enum class.

    output: Defines util functions to output dot files and util
        classes to store, represent detector results.
"""



from typing import Optional, List, Tuple, Union, TYPE_CHECKING

from tealer.teal.instructions.instructions import (
    PushInt,
    Instruction,
    Int,
    Byte,
    PushBytes,
    IntcInstruction,
    BytecInstruction,
)

from tealer.exceptions import TealerException

if TYPE_CHECKING:
    from tealer.teal.basic_blocks import BasicBlock
    from tealer.teal.functions import Function


def is_int_push_ins(ins: Instruction) -> Tuple[bool, Optional[Union[int, str]]]:
    """Return true if :ins: pushes a int literal on to the stack

    Args:
        ins: An instruction.

    Returns:
        Returns "pushes_int" and "value".

        pushes_int is True if the instruction pushes an uint64 value during execution.
        Otherwise, it is False.

        value is always None if pushes_int is False.
        value is None when pushes_int is True but the Tealer cannot compute the pushed value.
        value will be not None if Tealer can compute it.

        value will be a string if it is a named constant. The named constants are converted to
        integers by the assembler. Tealer directly returns the string instead of finding the compiled
        integer.

        value will be an int if the value is an integer in the Teal code.

    Raises:
        TealerException: Raises error if ins.bb or ins.bb.teal are not initialized properly.
    """
    if isinstance(ins, Int) or isinstance(  # pylint: disable=consider-merging-isinstance
        ins, PushInt
    ):
        return True, ins.value
    if isinstance(ins, IntcInstruction):
        if not ins.bb or not ins.bb.teal:
            # TODO: move the check to respective class property
            raise TealerException("Block or teal property is not set")
        teal = ins.bb.teal
        is_known, value = teal.get_int_constant(ins.index)
        if is_known:
            return True, value
        return True, None
    return False, None


def is_byte_push_ins(ins: Instruction) -> Tuple[bool, Optional[str]]:
    """Return true if :ins: pushes a byte literal on to the stack.

    Args:
        ins: An instruction.

    Returns:
        Returns "pushes_byte_value" and "value".

        pushes_byte_value is True if the instruction pushes an uint64 value during execution.
        Otherwise, it is False.

        value is always None if pushes_byte_value is False.
        value is None when pushes_byte_value is True but the Tealer cannot compute the pushed value.
        value will be the string if Tealer can compute it.

    Raises:
        TealerException: Raises error if ins.bb or ins.bb.teal are not initialized properly.
    """
    if isinstance(ins, (Byte, PushBytes)):
        return True, ins.value
    if isinstance(ins, BytecInstruction):
        if not ins.bb or not ins.bb.teal:
            # TODO: move the check to respective class property
            raise TealerException("Block or teal property is not set")
        teal = ins.bb.teal
        is_known, value = teal.get_byte_constant(ins.index)
        if is_known:
            return True, value
        return True, None
    return False, None


def next_blocks_global(function: "Function", block: "BasicBlock") -> List["BasicBlock"]:
    """Return basic blocks next to this block in the global CFG.

    global CFG is the single CFG representing the entire contract with callsub blocks connected
    to the subroutine entry blocks and retsub blocks of the subroutine connected to return point block.

    Args:
        function: The function
        block: A basic block.

    Returns:
        Returns the next basic blocks of :block: in the global CFG. The global CFG is the CFG where
        callsub_blocks are connected to the subroutine entry blocks.
    """
    if block.is_retsub_block:
        return function.return_point_blocks(block.subroutine)
    if block.is_callsub_block:
        return [block.called_subroutine.entry]
    return block.next


def prev_blocks_global(function: "Function", block: "BasicBlock") -> List["BasicBlock"]:
    """Return basic blocks previous to this block in the global CFG.

    global CFG is the single CFG representing the entire contract with callsub blocks connected
    to the subroutine entry blocks and retsub blocks of the subroutine connected to return point block.

    Args:
        function: The function
        block: A basic block.

    Returns:
        Returns the previous basic blocks of :block: in the global CFG. The global CFG is the CFG where
        callsub_blocks are connected to the subroutine entry blocks.
    """
    assert block.teal is not None
    if block == block.subroutine.entry:
        # if the block is the entry of the subroutine, return all blocks calling the subroutine
        if block.subroutine != function.main:
            return function.caller_blocks(block.subroutine)
        # the block is the main entry block of the contract
        return []
    if block.is_sub_return_point:
        # if the block is the return point of the subroutine, return all retsub blocks of the subroutine
        return block.callsub_block.called_subroutine.retsub_blocks
    # if its a normal block return previous blocks in the CFG.
    return block.prev


# TODO: Rename this to is_leaf_block_global
def leaf_block_global(block: "BasicBlock") -> bool:
    """Return True if block is a leaf block in the global CFG.

    global CFG is the single CFG representing the entire contract with callsub blocks connected
    to the subroutine entry blocks and retsub blocks of the subroutine connected to return point block.

    Args:
        block: A basic block.

    Returns:
        Returns True if :block: is a leaf block in the global CFG. otherwise, False.
    """
    # Retsub blocks will not have any next blocks but are not considered as leaf blocks in global CFG.
    # Callsub block will have a next block, which is executed after executing the subroutine. However, when
    # Callsub block is the last block in the source code and the subroutine exits the program using return instruction,
    # Then callsub block will not have a next block but it is not considered a leaf block in global CFG.
    return len(block.next) == 0 and not block.is_retsub_block and not block.is_callsub_block



from typing import Any, Dict
import pytest

from tealer.teal.instructions.parse_app_params_field import APP_PARAMS_FIELD_TXT_TO_OBJECT
from tealer.teal.instructions.parse_asset_holding_field import ASSET_HOLDING_FIELD_TXT_TO_OBJECT
from tealer.teal.instructions.parse_asset_params_field import ASSET_PARAMS_FIELD_TXT_TO_OBJECT
from tealer.teal.instructions.parse_global_field import GLOBAL_FIELD_TXT_TO_OBJECT
from tealer.teal.instructions.parse_transaction_field import TX_FIELD_TXT_TO_OBJECT
from tealer.teal.instructions import instructions

INSTRUCTION_TXT_TO_OBJECT = {
    "err": instructions.Err,
    "assert": instructions.Assert,
    "gaids": instructions.Gaids,
    "loads": instructions.Loads,
    "stores": instructions.Stores,
    "swap": instructions.Swap,
    "getbit": instructions.GetBit,
    "setbit": instructions.SetBit,
    "getbyte": instructions.GetByte,
    "setbyte": instructions.SetByte,
    "extract3": instructions.Extract3,
    "extract_uint16": instructions.Extract_uint16,
    "sha256": instructions.Sha256,
    "sha512_256": instructions.Sha512_256,
    "retsub": instructions.Retsub,
}

ALL_TESTS = [
    APP_PARAMS_FIELD_TXT_TO_OBJECT,
    ASSET_HOLDING_FIELD_TXT_TO_OBJECT,
    ASSET_PARAMS_FIELD_TXT_TO_OBJECT,
    GLOBAL_FIELD_TXT_TO_OBJECT,
    TX_FIELD_TXT_TO_OBJECT,
    INSTRUCTION_TXT_TO_OBJECT,
]


@pytest.mark.parametrize("test", ALL_TESTS)  # type: ignore
def test_string_repr(test: Dict[str, Any]) -> None:
    for target, obj in test.items():
        assert str(obj()) == target



"""
Helpers to fetch contract from the Algorand network using indexer.

## Test cases(Mainnet)

App id:
    - Invalid app id: 1
    - Valid app id: 991196662

Transaction id:
    - normal_txn (not signed with logic-sig): METJ6SNMBLTBVBLADEQFY6Y6CVPI4ZMSF23HVTY2RK7R4OCO6QDQ
    - delegate sig signed with single key: VFFVCSPFBK36ZN2FPDPDI6PFUO44OLZPBU6XPGOIGSR5L2XK6HFQ
    - delegate sig signed with multiple key(multi-sig): NGJOVS7SEEWRIKZNA5WL7FRENXXPUKSFPKJKM5QBYUOM7UZ2SG5Q
    - signed with contract account's sig: ZICHTBTBC6H3CFOGREUPV3G74WRYDM3ZD7W56VMRMCNXOQHP3AYQ

Contract Accounts:
    - Invalid address:  AOOIJDOIJRPZMCPJQWPIJNDISC2XEOZBDN2SNCONSOSDJX27652DNU2HDI
    - Contract account: D6DDUCPSDWD4IA2L3FL4ZC6NVKXGQYOU4LJQMHKV6NVCYHRURSFKYPFN5M
    - Normal account: ZW3ISEHZUHPO7OZGMKLKIIMKVICOUDRCERI454I3DB2BH52HGLSO67W754

### BETANET:

App id:
    - Invalid app id: 1
    - Valid app id: 1496883334      # worked for this application
    - Valid app id: 1497219145      # valid application but is not found by the indexer.

Dropping the support for betanet for now. Having no support is better than something that only works some times.
And supporting betanet is not a priority.

### TESTNET

App id:
    - Invalid app id: 1
    - Valid app id: 115884263
"""

import base64
import json

import requests
from algosdk.v2client.indexer import IndexerClient
from algosdk.error import IndexerHTTPError

from tealer.exceptions import TealerException


def disassemble_using_algo_explorer(code: str) -> str:
    """Disassemble the AVM bytecode to teal using Algorand node v2 api.

    Args:
        code: Compiled bytecode of the contract.

    Returns:
        Returns disassemble Teal code of :code:.

    Raises:
        TealerException: Raises error if the api returned any status code other than the
            success code (200).
    """
    base = "https://node.algoexplorerapi.io/"
    url = "v2/teal/disassemble"
    headers = {"Content-Type": "application/x-binary"}
    code_bin = base64.b64decode(code)
    r = requests.post(base + url, data=code_bin, headers=headers)
    if r.status_code != 200:
        raise TealerException(f'Unable to disassemble contract: "{code}"')
    disassembly = json.loads(r.text)["result"]
    return disassembly


def get_indexer(network: str) -> IndexerClient:
    main_net_base_url = "https://algoindexer.algoexplorerapi.io/"
    test_net_base_url = "https://algoindexer.testnet.algoexplorerapi.io/"
    # beta_net_base_url = "https://algoindexer.betanet.algoexplorerapi.io/"
    if network.lower() == "testnet":
        indexer_base_url = test_net_base_url
        print('Network: "TESTNET"')
    # Betannet is not supported
    # elif network == "betanet":
    #     indexer_base_url = beta_net_base_url
    #     print('Network: "BETANET"')
    elif network.lower() == "mainnet":
        indexer_base_url = main_net_base_url
        print('Network: "MAINET"')
    else:
        raise TealerException(f'Network "{network}" is not supported')

    return IndexerClient(indexer_token="", indexer_address=indexer_base_url)  # type: ignore


def get_application_using_app_id(network: str, app_id: int) -> str:
    """Downloads approval program using the application id

    Args:
        network: Algorand network identifier.
        app_id: Application Id.

    Returns:
        Returns disassembled Teal code of the :app_id: application fetched from the :network:.

    Raises:
        TealerException: Raises error if the function cannot find the application :app_id: in the
            :network:.
    """
    try:
        response = get_indexer(network).applications(application_id=app_id)  # type: ignore
    except IndexerHTTPError as e:
        raise TealerException(f'Unable to fetch application "{app_id}"\n{e}') from e
    approval = response["application"]["params"]["approval-program"]
    # clear = response["application"]["params"]["clear-state-program"]
    return disassemble_using_algo_explorer(approval)


def logic_sig_from_contract_account(network: str, account_address: str) -> str:
    """Find logic sig of a contract account.

    Logic sig of a contract account is not part of the account and is not stored in the
    blockchain state.

    In order to find the logic sig, find a transaction from the contract account and
    get the logic sig from that.

    Args:
        network: Algorand network identifier.
        account_address: An Algorand account address derived from a LogicSig.

    Returns:
        Returns disassembled Teal code of the contract controlling the contract account :account_address:.
        Contract is searched in the :network:.

    Raises:
        TealerException: Raises error if the function cannot find a transaction from the :account_address: in the
            :network:. Happens if the account_address is invalid, account has not made any transactions or the
            search has timed out.
    """
    try:
        results = get_indexer(network).search_transactions(  # type: ignore
            limit=1, sig_type="lsig", address=account_address, address_role="sender"
        )
    except IndexerHTTPError as e:
        # Happens if account_address is invalid.
        # or The search has timed-out.
        raise TealerException(
            f'Could not find logic signature for contract account "{account_address}"\n{e}'
        ) from e
    transactions = results["transactions"]
    if len(transactions) == 0:
        # Search has finished and didn't find any transactions.
        # -> Either account_address is not a contract account Or
        # -> There is not a single transaction from the account.
        raise TealerException(
            f'Account "{account_address}" is not a contract account.\n'
            "Note: Account should have made submitted atleast one transaction for tealer to find the contract"
        )
    signature = transactions[0]["signature"]["logicsig"]
    if "multi-signature" in signature or "signature" in signature:
        # delegate signature, not a contract account
        raise TealerException(f'Account "{account_address}" is not a contract account.')
    logic_sig = signature["logic"]
    return disassemble_using_algo_explorer(logic_sig)


def logic_sig_from_txn_id(network: str, txn_id: str) -> str:
    """Fetch logic-sig used to sign `txn_id` transaction

    Args:
        network: Algorand network identifier.
        txn_id: Transaction id of a transaction on :network:, the transaction is signed by a logic-sig.

    Returns:
        Returns disassembled Teal code of the contract that signed the transaction :txn_id:.
        The transaction is searched in the :network:.

    Raises:
        TealerException: Raises error if the function cannot find the transaction :txn_id: in the
            :network:
    """
    try:
        txn_info = get_indexer(network).transaction(txn_id)  # type: ignore
    except IndexerHTTPError as e:
        # will happen if txn_id is not a valid base32 string(irrespective of padding)
        # or There's no transaction with the given txn_id
        raise TealerException(f'Invalid transaction id: "{txn_id}"\n{e}') from e

    signature = txn_info["transaction"]["signature"]
    if "logicsig" not in signature:
        raise TealerException(f'Transaction "{txn_id}" is not signed using a logic signature.')
    logic_sig = signature["logicsig"]["logic"]
    return disassemble_using_algo_explorer(logic_sig)



from typing import List, Type
from prettytable import PrettyTable

from tealer.detectors.abstract_detector import (
    DETECTOR_TYPE_TXT,
    AbstractDetector,
    classification_txt,
)
from tealer.printers.abstract_printer import AbstractPrinter


def _sort_detector_classes(
    detector_classes: List[Type[AbstractDetector]],
) -> List[Type[AbstractDetector]]:
    # Sort by type, impact, confidence, name and description
    return sorted(
        detector_classes,
        key=lambda element: (element.TYPE, element.IMPACT, element.CONFIDENCE, element.NAME),
    )


def output_detectors(detector_classes: List[Type[AbstractDetector]]) -> None:
    """Print information of given detectors in the form of a table.

    Along with the name of detector, additional information such as what
    does detector tries to find/detect, type of contracts the detector is
    supposed to work on, impact(severity) of the issues detected by it,
    confidence(precision) of the detector are displayed. All this information
    is displayed in a table with appropriate columns. The detectors(rows)
    will be ordered based on the type, impact, confidence such that
    high impact, confidence detectors are displayed first.

    Args:
        detector_classes: list of detector classes whose information
            will be displayed by this function.
    """

    detector_classes = _sort_detector_classes(detector_classes)

    detectors_list = []
    for detector in detector_classes:
        name = detector.NAME
        description = detector.DESCRIPTION
        detector_type = DETECTOR_TYPE_TXT[detector.TYPE]
        detector_impact = detector.IMPACT
        detector_confidence = detector.CONFIDENCE
        detectors_list.append(
            (name, description, detector_type, detector_impact, detector_confidence)
        )
    table = PrettyTable(["Num", "Check", "What it Detects", "Type", "Impact", "Confidence"])

    idx = 1
    for (name, description, detector_type, impact, confidence) in detectors_list:
        table.add_row(
            [
                idx,
                name,
                description,
                detector_type,
                classification_txt[impact],
                classification_txt[confidence],
            ]
        )
        idx = idx + 1
    print(table)


def output_printers(printer_classes: List[Type[AbstractPrinter]]) -> None:
    """Print information of given printers in the form of a table.

    The name and the description of each printer is displayed. All
    printers are displayed in the form of a table with each row containing
    information about a single printer.

    Args:
        printer_classes: list of printer classes whose information
            will be displayed by this function.
    """

    printers_list = []
    for printer in printer_classes:
        name = printer.NAME
        printer_help = printer.HELP
        printers_list.append((name, printer_help))

    table = PrettyTable(["Num", "Printer", "What it Does"])

    printers_list = sorted(printers_list, key=lambda element: (element[0]))

    for (idx, (name, printer_help)) in enumerate(printers_list, start=1):
        table.add_row([idx, name, printer_help])

    print(table)


# pylint: disable=too-many-locals
def output_to_markdown(
    detector_classes: List[Type[AbstractDetector]],
    printer_classes: List[Type[AbstractPrinter]],
    filter_wiki: str,
) -> None:
    """Print information of detectors and printers in form of markdown table.

    Useful to automatically generate formatted description for available detectors and printers.
    Used to fill README.

    Args:
        detector_classes: List of available detectors.
        printer_classes: List of available printers.
        filter_wiki: Used to filter listed detectors based on NAME. A detector is listed if
            and only if filter_wiki is in its NAME.
    """

    def extract_help(cls: Type[AbstractDetector]) -> str:
        if cls.WIKI_URL == "":
            return cls.DESCRIPTION
        return f"[{cls.DESCRIPTION}]({cls.WIKI_URL})"

    detector_classes = _sort_detector_classes(detector_classes)

    detectors_list = []
    print(f"filter_wiki = {filter_wiki}")
    for detector in detector_classes:
        if not filter_wiki in detector.NAME:
            continue
        name = detector.NAME
        description = extract_help(detector)
        detector_type = DETECTOR_TYPE_TXT[detector.TYPE]
        detector_impact = detector.IMPACT
        detector_confidence = detector.CONFIDENCE
        detectors_list.append(
            (name, description, detector_type, detector_impact, detector_confidence)
        )

    idx = 1
    for (name, description, detector_type, impact, confidence) in detectors_list:
        print(
            f"{idx} | `{name}` | {description} | {detector_type} | {classification_txt[impact]} | {classification_txt[confidence]}"
        )
        idx = idx + 1

    print()
    printers_list = []
    for printer in printer_classes:
        argument = printer.NAME
        help_info = printer.HELP
        printers_list.append((argument, help_info))

    printers_list = sorted(printers_list, key=lambda element: (element[0]))
    idx = 1
    for (argument, help_info) in printers_list:
        print(f"{idx} | `{argument}` | {help_info}")
        idx = idx + 1


def output_wiki(detector_classes: List[Type[AbstractDetector]], filter_wiki: str) -> None:
    """Generate dectector documentation for github wiki.

    Args:
        detector_classes: List of detector classes.
        filter_wiki: filter string used to filter the detector classes. if this string
            is in a detector's name, then that detector is not included in the list.
    """
    detectors_list = _sort_detector_classes(detector_classes)

    for detector in detectors_list:
        if not filter_wiki in detector.NAME:
            continue
        check = detector.NAME
        applicable_to = DETECTOR_TYPE_TXT[detector.TYPE]
        impact = classification_txt[detector.IMPACT]
        confidence = classification_txt[detector.CONFIDENCE]
        title = detector.WIKI_TITLE
        description = detector.WIKI_DESCRIPTION
        exploit_scenario = detector.WIKI_EXPLOIT_SCENARIO
        recommendation = detector.WIKI_RECOMMENDATION

        print(f"\n## {title}")
        print("### Configuration")
        print(f"* Check: `{check}`")
        print(f"* Applicable to: `{applicable_to}`")
        print(f"* Severity: `{impact}`")
        print(f"* Confidence: `{confidence}`")
        print("\n### Description")
        print(description)
        if exploit_scenario:
            print("\n### Exploit Scenario:")
            print(exploit_scenario)
        print("\n### Recommendation")
        print(recommendation)



from typing import List, Tuple

from tealer.teal.basic_blocks import BasicBlock
from tealer.teal.instructions.instructions import Instruction


def cmp_instructions(ins1: Instruction, ins2: Instruction) -> bool:
    return str(ins1) == str(ins2)


def order_basic_blocks(bbs: List[BasicBlock]) -> List[BasicBlock]:
    key = lambda bb: bb.entry_instr.line
    return sorted(bbs, key=key)


def cmp_basic_blocks(b1: BasicBlock, b2: BasicBlock) -> bool:
    if len(b1.instructions) != len(b2.instructions):
        print(
            f"number of instructions are different: {len(b1.instructions)}, {len(b2.instructions)}"
        )
        return False

    for ins1, ins2 in zip(b1.instructions, b2.instructions):
        if not cmp_instructions(ins1, ins2):
            print(f"instructions comparison faile: {ins1}, {ins2}")
            return False

    return True


def cmp_cfg(bbs1: List[BasicBlock], bbs2: List[BasicBlock]) -> bool:
    if len(bbs1) != len(bbs2):
        print(f"number of blocks are different: {len(bbs1)}, {len(bbs2)}")
        return False
    bbs1, bbs2 = map(order_basic_blocks, (bbs1, bbs2))

    for bb1, bb2 in zip(bbs1, bbs2):
        print(f"comparing bb1: {bb1.idx}, bb2: {bb2.idx}\nbb1:\n{bb1}\nbb2:\n{bb2}\n")
        if not cmp_basic_blocks(bb1, bb2):
            return False

        if len(bb1.prev) != len(bb2.prev):
            print(f"number of prev blocks are different: {len(bb1.prev)}, {len(bb2.prev)}")
            print(bb1.prev)
            print(bb2.prev)
            return False

        if len(bb1.next) != len(bb2.next):
            print(f"number of next blocks are different: {len(bb1.next)}, {len(bb2.next)}")
            print(bb1.next)
            print(bb2.next)
            return False

        bbs_prev_1, bbs_prev_2 = map(order_basic_blocks, (bb1.prev, bb2.prev))
        for bbp1, bbp2 in zip(bbs_prev_1, bbs_prev_2):
            if bbp1.entry_instr.line != bbp2.entry_instr.line:
                print("entry lines of previous blocks is different")
                print(bbp1.entry_instr.line, bbp2.entry_instr.line)
                return False

        bbs_next_1, bbs_next_2 = map(order_basic_blocks, (bb1.next, bb2.next))
        for bbn1, bbn2 in zip(bbs_next_1, bbs_next_2):
            if bbn1.entry_instr.line != bbn2.entry_instr.line:
                print("entry lines of next blocks is different")
                print(bbn1.entry_instr.line, bbn2.entry_instr.line)
                return False

    return True


def construct_cfg(
    ins_list: List[Instruction],
    ins_partitions: List[Tuple[int, int]],
    bbs_links: List[Tuple[int, int]],
) -> List[BasicBlock]:
    """cfg construction helper function.

    construct cfg using the list of instructions, basic block partitions
    and their relation.

    Args:
        ins_list: list of instructions of the program.
        ins_partitions (List[Tuple[int]]): list of tuple indicating starting and
                ending index into ins_list where ins_list[starting: ending] belong to a
                single basic_block.
        bbs_links (List[Tuple[int]]): links two basic blocks (b1, b2). indexes of basic block
                corresponds to index in ins_partition. order of indexes is important as this adds
                a directed edge from b1 to b2.

    Returns:
        List[BasicBlock]: returns a list of basic blocks representing the CFG.
    """
    for idx, ins in enumerate(ins_list, start=1):
        ins.line = idx

    bbs = []
    for (idx, (start, end)) in enumerate(ins_partitions):
        bb = BasicBlock()
        for ins in ins_list[start:end]:
            bb.add_instruction(ins)
        bb.idx = idx
        bbs.append(bb)

    for idx1, idx2 in bbs_links:
        bbs[idx1].add_next(bbs[idx2])
        bbs[idx2].add_prev(bbs[idx1])

    return bbs



from typing import TYPE_CHECKING, Optional, Dict, List

from tealer.utils.teal_enums import TransactionType

# from tealer.exceptions import TealerException

if TYPE_CHECKING:
    from tealer.teal.functions import Function

# pylint: disable=too-few-public-methods, too-many-instance-attributes
class Transaction:
    def __init__(self) -> None:
        self.type: TransactionType = TransactionType.Any
        self.has_logic_sig: bool = False
        self.logic_sig: Optional["Function"] = None
        self.application: Optional["Function"] = None
        self.absoulte_index: Optional[int] = None
        # t1.relative_indexes[-1] = t2 => t2.group_index() == t1.group_index() - 1
        self.relative_indexes: Dict[int, Transaction] = {}
        self.group_transaction: Optional[GroupTransaction] = None
        self.transacton_id: str = ""


# pylint: disable=too-few-public-methods
class GroupTransaction:
    def __init__(self) -> None:
        self.transactions: List[Transaction] = []
        self.absolute_indexes: Dict[int, Transaction] = {}
        # {t1: (t2, -2)} => t1.group_index() == t2.group_index() - 2
        self.group_relative_indexes: Dict[Transaction, Dict[Transaction, int]] = {}
        self.operation_name: str = ""


def fill_group_relative_indexes(group: "GroupTransaction") -> None:
    # group_relative_indexes: Dict[Transaction, Dict[Transaction, int]] = {}
    for txn in group.transactions:
        group.group_relative_indexes[txn] = {}

    for txn in group.transactions:
        for offset in txn.relative_indexes:
            other_txn = txn.relative_indexes[offset]
            # other_txn.group_index() = txn.group_index() + offset
            group.group_relative_indexes[other_txn][txn] = offset


# def fill_indexes(group: GroupTransaction):
#     """Calculate the absolute indexes and relative indexes using the indexes from the config.

#     """

#     for txn in group.transactions:
#         if txn.absoulte_index is not None:
#             for offset in txn.relative_indexes:
#                 other_txn = txn.relative_indexes[offset]
#                 other_txn_abs = txn.absoulte_index + offset
#                 if other_txn.absoulte_index is None:
#                     other_txn.absoulte_index = other_txn_abs
#                     group.absolute_indexes[other_txn_abs] = other_txn
#                 # else:
#                 #     if other_txn.absoulte_index != other_txn_abs:
#                         # raise TealerException("Invalid group configuration")
#         for offset in txn.relative_indexes:
#             other_txn = txn.relative_indexes[offset]
#             this_txn_offset = -1 * offset
#             if this_txn_offset not in other_txn.relative_indexes:
#                 other_txn.relative_indexes[this_txn_offset] = txn



"""Defines class to represent Teal contracts.

Teal class is used to store Control Flow Graph of the contract.
It comes with methods for running detectors, printers on the contract.
And also has properties for easy access of the parsed contract
information.

Classes:
    Teal: Class to represent a contract.
"""

from typing import List, TYPE_CHECKING, Tuple, Dict, Optional

from tealer.teal.basic_blocks import BasicBlock
from tealer.teal.subroutine import Subroutine
from tealer.teal.instructions.instructions import Instruction
from tealer.utils.teal_enums import ContractType, ExecutionMode

if TYPE_CHECKING:
    from tealer.teal.functions import Function


# disable pylint errors for now. errors should be resolved after completing refactoring.
class Teal:  # pylint: disable=too-many-instance-attributes,too-many-public-methods
    """Class to represent a teal contract.

    This class stores CFG, subroutines and other information of the
    contract. Also comes with easy api methods to run detectors, printers
    on the contract.

    Args:
        instructions: List of parsed instruction objects.
        bbs: List of basic blocks representing the Control Flow Graph(CFG)
            of the contract.
        version: Teal version this contract is written in.
        mode: Mode of the contract.
        subroutines: List of subroutines defined in the contract, will be
            empty for contracts written in Teal version 3 or less. Each
            subroutine is represented by the list of basic blocks part of
            that subroutine.
    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        version: int,
        mode: ExecutionMode,
        instructions: List[Instruction],
        bbs: List[BasicBlock],
        main: Subroutine,
        subroutines: Dict[str, Subroutine],
    ):
        self._version = version
        self._mode = mode
        self._int_constants: List[int] = []
        self._byte_constants: List[str] = []

        self._instructions = instructions
        self._bbs = bbs
        self._main = main
        self._subroutines = subroutines
        self._functions: Dict[str, "Function"] = {}

        self._contract_name: str = ""
        self._contract_type: ContractType = (
            ContractType.ApprovalProgram
            if mode == ExecutionMode.STATEFUL
            else ContractType.LogicSig
        )

    @property
    def instructions(self) -> List[Instruction]:
        """List of instructions of the contract

        Returns:
            List of instructions of the contract
        """
        return self._instructions

    @property
    def bbs(self) -> List[BasicBlock]:
        """List of all basic blocks of the contract

        Returns:
            List of all basic blocks of the contract
        """
        return self._bbs

    @property
    def version(self) -> int:
        """Teal version for this contract.

        Version of the contract is based on whether the first instruction of the
        contract is #pragma version instruction or not. if it is, then version will
        be the teal version specified or else version will be 1.

        Returns:
            Returns the contract version.
        """

        return self._version

    @version.setter
    def version(self, ver: int) -> None:
        self._version = ver

    @property
    def mode(self) -> ExecutionMode:
        """Execution mode of the contract: Stateless, Stateful or Any.

        Mode is determined based on the instructions of the contract. if there are
        any instructions that are only supported in one kind of contract then that
        will be the contract type. if there aren't any such type of instructions, then
        this will be "Any".

        Returns:
            Returns the execution mode this contract can be executed successfully.
        """

        return self._mode

    @mode.setter
    def mode(self, m: ExecutionMode) -> None:
        self._mode = m

    @property
    def contract_type(self) -> ContractType:
        """Type of the contract: ApprovalProgram, ClearStateProgram or LogicSig

        Returns:
            Returns the type of the contract
        """
        return self._contract_type

    @contract_type.setter
    def contract_type(self, contract_type: ContractType) -> None:
        self._contract_type = contract_type

    @property
    def main(self) -> "Subroutine":
        """Returns subroutine representing the contract entry-point

        Returns:
            Returns the subroutine object representing the non-subroutine blocks.
        """
        return self._main

    @property
    def subroutines(self) -> Dict[str, "Subroutine"]:
        """Returns dict of subroutine names and corresponding subroutine obj.

        Returns:
            Returns the subroutines in the contract. keys are the subroutine's name
            and the value is the subroutine.
        """
        return self._subroutines

    @property
    def subroutines_list(self) -> List["Subroutine"]:
        """Returns list of all contract's subroutines

        Returns:
            Returns the list of all subroutines in the contract.
        """
        return list(self._subroutines.values())

    @property
    def functions(self) -> Dict[str, "Function"]:
        return self._functions

    @functions.setter
    def functions(self, functions: Dict[str, "Function"]) -> None:
        self._functions = functions

    @property
    def functions_list(self) -> List["Function"]:
        return list(self._functions.values())

    @property
    def contract_name(self) -> str:
        return self._contract_name

    @contract_name.setter
    def contract_name(self, name: str) -> None:
        self._contract_name = name

    def subroutine(self, name: str) -> Optional["Subroutine"]:
        """Return subroutine with id/name `name`, return none if subroutine does not exist.

        Args:
            name: name of the subroutine.

        Returns:
            Returns the subroutine object given its name. Returns None if the subroutine is
            not in identified subroutines list.
        """
        return self._subroutines.get(name, None)

    def get_int_constant(self, index: int) -> Tuple[bool, int]:
        """Return int value stored by intcblock instruction

        Returns boolean and an int. Boolean indicates whether there's a value at that index
        and whether Tealer was able to identify that information.

        Cases:
            intcblock is only present in the entry block; Tealer is able to identify the confirmed and
            correct constants. As a result, Tealer can know the values of `intc_*` instructions.

            intcblock instruction occurs multiple times. Tealer cannot determine the exact value a `intc_`
            instruction refers to. This method returns `False` in this case.

            A `intc_*` instruction refers to a value that is not declared in the intcblock instruction. For example,
            intcblock only stores 3 values in the constant space and a intc instruction refers to 5th value. Execution
            will fail at runtime in this case. Return value will be `False` in this case as well.

        Args:
            index: The index into the int constants filled by the intcblock instruction.

        Returns:
            bool: True if Tealer was able to determine the value referred by that instruction or else False
            int: value referred by intc instruction at that index.
        """
        if len(self._int_constants) <= index:
            return False, 0
        return True, self._int_constants[index]

    def get_byte_constant(self, index: int) -> Tuple[bool, str]:
        """Return []byte value stored by bytecblock instruction

        Behavior is same as `get_int_constant`.

        Args:
            index: The index into the byte constants filled by the bytecblock instruction.

        Returns:
            bool: True if Tealer was able to determine the value referred by that instruction or else False
            int: value referred by bytec instruction at that index.
        """
        if len(self._byte_constants) <= index:
            return False, ""
        return True, self._byte_constants[index]

    def set_int_constants(self, int_constants: List[int]) -> None:
        self._int_constants = int_constants

    def set_byte_constants(self, byte_constants: List[str]) -> None:
        self._byte_constants = byte_constants



# pylint: skip-file
# mypy: ignore-errors
from typing import List
from pyteal import *  # pylint: disable=wildcard-import, unused-wildcard-import


from tealer.detectors.all_detectors import MissingGroupSize


@Subroutine(TealType.none)
def transfer_wrapped_algo(_receiver: Expr, _amount: Expr) -> Expr:
    return Pop(Bytes("Inner txn to transfer assets"))


@Subroutine(TealType.none)
def transfer_algo(_receiver: Expr, _amount: Expr) -> Expr:
    return Pop(Bytes("Inner txn to transfer Algos"))


def mint_wrapped_algo() -> Expr:
    validations = Assert(
        And(
            Gtxn[0].receiver() == Global.current_application_address(),
            Gtxn[0].type_enum() == TxnType.Payment,
        )
    )
    transfer_op = transfer_wrapped_algo(Txn.sender(), Gtxn[0].amount())
    return Seq([validations, transfer_op, Approve()])


def burn_wrapped_algo() -> Expr:
    validations = Assert(
        And(
            # Gtxn[Int(x)] => int x; gtxn ...
            Gtxn[Int(0)].receiver() == Global.current_application_address(),
            Gtxn[Int(0)].type_enum() == TxnType.AssetTransfer,
            Gtxn[Int(0)].xfer_asset() == Int(1337),
        )
    )
    transfer_op = transfer_algo(Txn.sender(), Gtxn[0].asset_amount())
    return Seq([validations, transfer_op, Approve()])


def mint_wrapped_algo_fixed() -> Expr:
    validations = Assert(
        And(
            # Gtxn[Int(x)] => int x; gtxn ...
            Gtxn[Int(0)].receiver() == Global.current_application_address(),
            Gtxn[0].type_enum() == TxnType.Payment,
            Global.group_size() == Int(2),
        )
    )
    transfer_op = transfer_wrapped_algo(Txn.sender(), Gtxn[0].amount())
    return Seq([validations, transfer_op, Approve()])


def burn_wrapped_algo_fixed() -> Expr:
    index = Txn.group_index() - Int(1)
    validations = Assert(
        And(
            Gtxn[index].receiver() == Global.current_application_address(),
            Gtxn[index].type_enum() == TxnType.AssetTransfer,
            Gtxn[index].xfer_asset() == Int(1337),
        )
    )
    transfer_op = transfer_algo(Txn.sender(), Gtxn[index].asset_amount())
    return Seq([validations, transfer_op, Approve()])


def wrapped_approval_program_1() -> Expr:
    return Cond(
        [Txn.application_args[0] == Bytes("mint"), mint_wrapped_algo()],
        [Txn.application_args[0] == Bytes("burn"), burn_wrapped_algo()],
    )


def wrapped_approval_program_2() -> Expr:
    return Cond(
        [Txn.application_args[0] == Bytes("mint"), mint_wrapped_algo_fixed()],
        [Txn.application_args[0] == Bytes("burn"), burn_wrapped_algo()],
    )


def wrapped_approval_program_3() -> Expr:
    return Cond(
        [Txn.application_args[0] == Bytes("mint"), mint_wrapped_algo()],
        [Txn.application_args[0] == Bytes("burn"), burn_wrapped_algo_fixed()],
    )


def wrapped_approval_program_4() -> Expr:
    return Cond(
        [Txn.application_args[0] == Bytes("mint"), mint_wrapped_algo_fixed()],
        [Txn.application_args[0] == Bytes("burn"), burn_wrapped_algo_fixed()],
    )


wrapped_approval_program_1_teal = compileTeal(
    wrapped_approval_program_1(),
    mode=Mode.Application,
    version=8,
)

wrapped_approval_program_1_vulnerable_paths: List[List[int]] = [
    [0, 1, 3, 8, 4],  # burn
    [0, 5, 7, 6],  # mint
]

wrapped_approval_program_2_teal = compileTeal(
    wrapped_approval_program_2(),
    mode=Mode.Application,
    version=8,
)

wrapped_approval_program_2_vulnerable_paths: List[List[int]] = [
    [0, 1, 3, 8, 4],
]

wrapped_approval_program_3_teal = compileTeal(
    wrapped_approval_program_3(),
    mode=Mode.Application,
    version=8,
)

wrapped_approval_program_3_vulnerable_paths: List[List[int]] = [
    [0, 5, 7, 6],
]

wrapped_approval_program_4_teal = compileTeal(
    wrapped_approval_program_4(),
    mode=Mode.Application,
    version=8,
)

wrapped_approval_program_4_vulnerable_paths: List[List[int]] = []


group_size_tests_pyteal = [
    (
        wrapped_approval_program_1_teal,
        MissingGroupSize,
        wrapped_approval_program_1_vulnerable_paths,
    ),
    (
        wrapped_approval_program_2_teal,
        MissingGroupSize,
        wrapped_approval_program_2_vulnerable_paths,
    ),
    (
        wrapped_approval_program_3_teal,
        MissingGroupSize,
        wrapped_approval_program_3_vulnerable_paths,
    ),
    (
        wrapped_approval_program_4_teal,
        MissingGroupSize,
        wrapped_approval_program_4_vulnerable_paths,
    ),
]

if __name__ == "__main__":
    print(wrapped_approval_program_3_teal)



import sys
from dataclasses import dataclass
from pathlib import Path
from typing import List, Dict, Any, Optional
import yaml

from tealer.utils.teal_enums import TransactionType


GROUP_CONFIG_CONTRACT_TYPES = [
    "LogicSig",
    "ApprovalProgram",
    "ClearStateProgram",
]

USER_CONFIG_TRANSACTION_TYPES = {
    "pay": TransactionType.Pay,
    "keyreg": TransactionType.KeyReg,
    "acfg": TransactionType.Acfg,
    "axfer": TransactionType.Axfer,
    "afrz": TransactionType.Afrz,
    "appl": TransactionType.Appl,
    "txn": TransactionType.Any,
}


class InvalidGroupConfiguration(Exception):
    pass


def check_fields_are_present(required_fields: List[str], config: Dict[str, Any]) -> List[str]:
    absent_fields: List[str] = []
    for field in required_fields:
        if field not in config:
            absent_fields.append(field)
    return absent_fields


@dataclass
class GroupConfigFunction:
    name: str
    dispatch_path: List[str]

    @staticmethod
    def from_yaml(function: Dict[str, Any]) -> "GroupConfigFunction":
        if "name" not in function:
            raise InvalidGroupConfiguration("function name is not given")
        name: str = function["name"]
        absent_fields = check_fields_are_present(["dispatch_path"], function)
        if absent_fields:
            raise InvalidGroupConfiguration(
                f'Function name: {name}\n\nFollowing Required fields are absent: {", ".join(absent_fields)}'
            )

        dispatch_path: List[str] = function["dispatch_path"]
        for block_id in dispatch_path:
            if not block_id.startswith("B") or not block_id[1:].isdigit():
                raise InvalidGroupConfiguration(
                    f"Function name: {name}\nIncorrect block id in dispatch path: {dispatch_path}"
                )

        return GroupConfigFunction(name, dispatch_path)

    def to_yaml(self) -> Dict[str, Any]:
        return {"name": self.name, "dispatch_path": self.dispatch_path}


@dataclass
class GroupConfigContract:
    name: str
    file_path: Path
    contract_type: str
    version: int
    subroutines: List[str]
    functions: List[GroupConfigFunction]

    @staticmethod
    def from_yaml(contract: Dict[str, Any]) -> "GroupConfigContract":
        if "name" not in contract:
            raise InvalidGroupConfiguration("contract name is not given")
        name: str = contract["name"]

        required_fields = ["file_path", "type", "version", "subroutines", "functions"]
        absent_fields: List[str] = []
        for field in required_fields:
            if field not in contract:
                absent_fields.append(field)
        if absent_fields:
            raise InvalidGroupConfiguration(
                f'Contract name: {name}\n\nFollowing Required fields are absent: {", ".join(absent_fields)}'
            )

        file_path: Path = Path(contract["file_path"])
        contract_type: str = contract["type"]
        version: int = contract["version"]
        subroutines: List[str] = contract["subroutines"]

        if contract_type not in GROUP_CONFIG_CONTRACT_TYPES:
            raise InvalidGroupConfiguration(
                f"Contract name: {name}\n Invalid contract type: {contract_type}"
            )

        functions: List[Dict[str, Any]] = contract["functions"]
        parsed_functions: List[GroupConfigFunction] = []

        for function in functions:
            try:
                parsed_functions.append(GroupConfigFunction.from_yaml(function))
            except InvalidGroupConfiguration as err:
                # pylint: disable=raise-missing-from
                raise InvalidGroupConfiguration(f"Contract name: {name}\n{err}")

        return GroupConfigContract(
            name, file_path, contract_type, version, subroutines, parsed_functions
        )

    def to_yaml(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "file_path": self.file_path,
            "type": self.contract_type,
            "version": self.version,
            "subroutines": self.subroutines,
            "functions": [function.to_yaml() for function in self.functions],
        }


@dataclass
class GroupConfigFunctionCall:
    contract: str
    function: str

    @staticmethod
    def from_yaml(function_call: Dict[str, Any]) -> "GroupConfigFunctionCall":
        absent_fields = check_fields_are_present(["contract", "function"], function_call)
        if absent_fields:
            raise InvalidGroupConfiguration(
                f'Function call:\n\nFollowing Required fields are absent: {", ".join(absent_fields)}'
            )
        contract = function_call["contract"]
        function = function_call["function"]

        return GroupConfigFunctionCall(contract, function)

    def to_yaml(self) -> Dict[str, Any]:
        return {
            "contract": self.contract,
            "function": self.function,
        }


@dataclass
class GroupConfigTransaction:
    txn_id: str
    txn_type: str
    application: Optional[GroupConfigFunctionCall] = None
    has_logic_sig: Optional[bool] = None
    logic_sig: Optional[GroupConfigFunctionCall] = None
    absolute_index: Optional[int] = None
    relative_indexes: Optional[Dict[str, int]] = None

    @staticmethod
    def from_yaml(transaction: Dict[str, Any]) -> "GroupConfigTransaction":
        absent_fields = check_fields_are_present(["txn_id", "txn_type"], transaction)
        if absent_fields:
            raise InvalidGroupConfiguration(
                f'Transaction:\n\nFollowing Required fields are absent: {", ".join(absent_fields)}'
            )

        txn_id = transaction["txn_id"]
        txn_type = transaction["txn_type"]
        if txn_type not in USER_CONFIG_TRANSACTION_TYPES:
            raise InvalidGroupConfiguration(
                f"Transaction: Unknown transaction type {txn_type} of transaction {txn_id}"
            )
        application = transaction.get("application")
        has_logic_sig = transaction.get("has_logic_sig")
        logic_sig = transaction.get("logic_sig")
        absolute_index = transaction.get("absolute_index")
        relative_indexes = transaction.get("relative_indexes")

        if application is not None:
            application = GroupConfigFunctionCall.from_yaml(application)

        if logic_sig is not None:
            logic_sig = GroupConfigFunctionCall.from_yaml(logic_sig)

        if relative_indexes is not None:
            parsed_relative_indexes: Dict[str, int] = {}
            for relative_index in relative_indexes:
                absent_fields = check_fields_are_present(["other_txn_id", "offset"], relative_index)
                if absent_fields:
                    raise InvalidGroupConfiguration(
                        f'Transaction: {txn_id}\n\nFollowing Required fields are absent in relative_indexes: {", ".join(absent_fields)}'
                    )
                parsed_relative_indexes[relative_index["other_txn_id"]] = relative_index["offset"]
            relative_indexes = parsed_relative_indexes

        return GroupConfigTransaction(
            txn_id,
            txn_type,
            application,
            has_logic_sig,
            logic_sig,
            absolute_index,
            relative_indexes,
        )

    def to_yaml(self) -> Dict[str, Any]:
        output: Dict[str, Any] = {
            "txn_id": self.txn_id,
            "txn_type": self.txn_type,
        }
        if self.application is not None:
            output["application"] = self.application.to_yaml()
        if self.logic_sig is not None:
            output["logic_sig"] = self.logic_sig.to_yaml()
        if self.has_logic_sig is not None:
            output["has_logic_sig"] = self.has_logic_sig
        if self.absolute_index is not None:
            output["absolute_index"] = self.absolute_index
        if self.relative_indexes is not None:
            relative_indexes = []
            for txn_id, offset in self.relative_indexes.items():
                relative_indexes.append(
                    {
                        "other_txn_id": txn_id,
                        "offset": offset,
                    }
                )
            output["relative_indexes"] = relative_indexes

        return output


@dataclass
class GroupConfigGroup:
    operation: str
    transactions: List[GroupConfigTransaction]

    @staticmethod
    def from_yaml(group: Dict[str, Any]) -> "GroupConfigGroup":
        absent_fields = check_fields_are_present(["operation", "transactions"], group)
        if absent_fields:
            raise InvalidGroupConfiguration(
                f'Group:\n\nFollowing Required fields are absent: {", ".join(absent_fields)}'
            )

        operation = group["operation"]
        parsed_transactions = []
        for transaction in group["transactions"]:
            parsed_transactions.append(GroupConfigTransaction.from_yaml(transaction))

        return GroupConfigGroup(operation, parsed_transactions)

    def to_yaml(self) -> Dict[str, Any]:
        return {
            "operation": self.operation,
            "transactions": [transaction.to_yaml() for transaction in self.transactions],
        }


@dataclass
class GroupConfig:
    name: str
    contracts: List[GroupConfigContract]
    groups: List[GroupConfigGroup]

    @staticmethod
    def from_yaml(config: Dict[str, Any]) -> "GroupConfig":
        absent_fields = check_fields_are_present(["name", "contracts", "groups"], config)
        if absent_fields:
            raise InvalidGroupConfiguration(
                f'Config:\n\nFollowing Required fields are absent: {", ".join(absent_fields)}'
            )
        name = config["name"]
        contracts = []
        for contract in config["contracts"]:
            contracts.append(GroupConfigContract.from_yaml(contract))
        groups = []
        for group in config["groups"]:
            groups.append(GroupConfigGroup.from_yaml(group))

        return GroupConfig(name, contracts, groups)

    def to_yaml(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "contracts": [contract.to_yaml() for contract in self.contracts],
            "groups": [group.to_yaml() for group in self.groups],
        }


def write_to_yaml_file(file: Path, output: Dict[str, Any]) -> None:
    with open(file, "w", encoding="utf-8") as f:
        # yaml = YAML()
        # yaml.indent(sequence=4, offset=2)
        # yaml.dump(output, f)
        yaml.safe_dump(output, f, sort_keys=False)


def read_config_from_file(file: Path) -> GroupConfig:
    with open(file, encoding="utf-8") as f:
        d = yaml.safe_load(f.read())
    try:
        parsed_config = GroupConfig.from_yaml(d)
        # given file_path is relative path of the contract from config file.
        for contract in parsed_config.contracts:
            contract.file_path = file.parent / contract.file_path
        return parsed_config
    except InvalidGroupConfiguration as err:
        print(f"\nInvalidGroupConfiguration:\n\n{err}")
        print(err)
        sys.exit(1)


# if __name__ == "__main__":
# filename = "tests/group_transactions/ans/ans_config.yaml"
# with open("tests/group_transactions/ans_config.yaml") as f:
# try:
# if 1:
# parsed_config = read_config_from_file(filename)
# write_to_yaml_file(Path("test.yaml"), parsed_config.to_yaml())
# except InvalidGroupConfiguration as err:
# print(f"\nInvalidGroupConfiguration:\n\n{err}")
# print(err)
# sys.exit(1)



from typing import List, Optional, Dict
from dataclasses import dataclass, field

from tealer.exceptions import TealerException
from tealer.utils.teal_enums import ALL_TRANSACTION_TYPES
from tealer.utils.algorand_constants import MAX_GROUP_SIZE, MAX_UINT64


@dataclass
class AddrFieldValue:
    any_addr: bool = True
    no_addr: bool = False
    possible_addr: List[str] = field(default_factory=list)


class BlockTransactionContext:  # pylint: disable=too-few-public-methods, too-many-instance-attributes

    _gtxn_at_index_context: Optional[List["BlockTransactionContext"]] = None
    _abs_context: Optional[List["BlockTransactionContext"]] = None
    _relative_context: Optional[Dict[int, "BlockTransactionContext"]] = None

    def __init__(self, tail: bool = False) -> None:
        if not tail:
            self._gtxn_at_index_context = [
                BlockTransactionContext(True) for _ in range(MAX_GROUP_SIZE)
            ]
            self._abs_context = [BlockTransactionContext(True) for _ in range(MAX_GROUP_SIZE)]
            self._relative_context = {
                offset: BlockTransactionContext(True)
                for offset in range(-(MAX_GROUP_SIZE - 1), MAX_GROUP_SIZE)
                if offset != 0
            }

        # set default values
        if tail:
            # information from gtxn {i} instructions.
            self.group_indices = []
            self.group_sizes = []
            self.is_gtxn_context = True
        else:
            self.group_sizes = list(range(1, MAX_GROUP_SIZE + 1))
            self.group_indices = list(range(0, MAX_GROUP_SIZE))
            self.is_gtxn_context = False
        self.transaction_types = list(ALL_TRANSACTION_TYPES)
        self.rekeyto: AddrFieldValue = AddrFieldValue()
        self.closeto: AddrFieldValue = AddrFieldValue()
        self.assetcloseto: AddrFieldValue = AddrFieldValue()
        self.sender: AddrFieldValue = AddrFieldValue()
        self.max_fee: int = MAX_UINT64
        self.max_fee_unknown: bool = False  # True if max possible fee is bounded and unknown

    def gtxn_context(self, txn_index: int) -> "BlockTransactionContext":
        """context information collected from gtxn {txn_index} field instructions

        Args:
            txn_index: Transaction index.

        Returns:
            context information collected from "gtxn {txn_index} field" instructions.
            The information is of the txn executing this contract when it's index in the group
            is :txn_index:

        Raises:
            TealerException: Raises error if gtxn_context of a gtxn_context is accessed or if the
                transaction index is greater than MAX_GROUP_SIZE.
        """
        if self._gtxn_at_index_context is None:
            raise TealerException()
        if txn_index >= MAX_GROUP_SIZE:
            raise TealerException()
        return self._gtxn_at_index_context[txn_index]

    def absolute_context(self, txn_index: int) -> "BlockTransactionContext":
        """context information of the transaction at `txn_index`

        Args:
            txn_index: Transaction index

        Returns:
            The context information of the transaction at the give absolute index.

        Raises:
            TealerException: Raises error if abs_context of a tail context is accessed or if the transaction
                index is greater than MAX_GROUP_SIZE.
        """
        if self._abs_context is None:
            raise TealerException()
        if txn_index >= MAX_GROUP_SIZE:
            raise TealerException()
        return self._abs_context[txn_index]

    def relative_context(self, offset: int) -> "BlockTransactionContext":
        """context information of the transaction at `offset` from the current transaction

        Args:
            offset: Transaction offset

        Returns:
            The context information of the transaction at the offset from the executing this contract

        Raises:
            TealerException: Raises error if r of a tail context is accessed or if the offset is not valid
        """
        if self._relative_context is None:
            raise TealerException()
        if offset not in self._relative_context:
            raise TealerException()
        return self._relative_context[offset]



"""Parser for transaction fields.

Each transaction field is represented as a class. Parsing the field
is creating the class instance representing the field given it's
string representation.

Most of the transaction fields doesn't have immediate arguments
and their string representation consists of single sequence of characters.
Few transaction fields are arrays and have single immediate argument
which is the index into the array. Array Transaction fields with single immediate
argument are parsed using ARRAY_TX_FIELD_TO_OBJECT. For other fields,
TX_FIELD_TXT_TO_OJECT is used as lookup.

Attributes:
    TX_FIELD_TXT_TO_OBJECT: Map(dict) from string representation
        of transaction field to the corresponding class.
    ARRAY_TX_FIELD_TXT_TO_OBJECT: Map(dict) from string representation
        of array transaction field to the corresponding class.
"""
from typing import Dict, Type
from tealer.teal.instructions import transaction_field

TX_FIELD_TXT_TO_OBJECT: Dict[str, Type[transaction_field.TransactionField]] = {
    "Sender": transaction_field.Sender,
    "Fee": transaction_field.Fee,
    "FirstValid": transaction_field.FirstValid,
    "FirstValidTime": transaction_field.FirstValidTime,
    "LastValid": transaction_field.LastValid,
    "Note": transaction_field.Note,
    "Lease": transaction_field.Lease,
    "Receiver": transaction_field.Receiver,
    "Amount": transaction_field.Amount,
    "CloseRemainderTo": transaction_field.CloseRemainderTo,
    "VotePK": transaction_field.VotePK,
    "SelectionPK": transaction_field.SelectionPK,
    "VoteFirst": transaction_field.VoteFirst,
    "VoteLast": transaction_field.VoteLast,
    "VoteKeyDilution": transaction_field.VoteKeyDilution,
    "Type": transaction_field.Type,
    "TypeEnum": transaction_field.TypeEnum,
    "XferAsset": transaction_field.XferAsset,
    "AssetAmount": transaction_field.AssetAmount,
    "AssetSender": transaction_field.AssetSender,
    "AssetReceiver": transaction_field.AssetReceiver,
    "AssetCloseTo": transaction_field.AssetCloseTo,
    "GroupIndex": transaction_field.GroupIndex,
    "TxID": transaction_field.TxID,
    "ApplicationID": transaction_field.ApplicationID,
    "OnCompletion": transaction_field.OnCompletion,
    "NumAppArgs": transaction_field.NumAppArgs,
    "NumAccounts": transaction_field.NumAccounts,
    "NumApplications": transaction_field.NumApplications,
    "NumAssets": transaction_field.NumAssets,
    "ApprovalProgram": transaction_field.ApprovalProgram,
    "ClearStateProgram": transaction_field.ClearStateProgram,
    "RekeyTo": transaction_field.RekeyTo,
    "ConfigAsset": transaction_field.ConfigAsset,
    "ConfigAssetTotal": transaction_field.ConfigAssetTotal,
    "ConfigAssetDecimals": transaction_field.ConfigAssetDecimals,
    "ConfigAssetDefaultFrozen": transaction_field.ConfigAssetDefaultFrozen,
    "ConfigAssetUnitName": transaction_field.ConfigAssetUnitName,
    "ConfigAssetName": transaction_field.ConfigAssetName,
    "ConfigAssetURL": transaction_field.ConfigAssetURL,
    "ConfigAssetMetadataHash": transaction_field.ConfigAssetMetadataHash,
    "ConfigAssetManager": transaction_field.ConfigAssetManager,
    "ConfigAssetReserve": transaction_field.ConfigAssetReserve,
    "ConfigAssetFreeze": transaction_field.ConfigAssetFreeze,
    "ConfigAssetClawback": transaction_field.ConfigAssetClawback,
    "FreezeAsset": transaction_field.FreezeAsset,
    "FreezeAssetAccount": transaction_field.FreezeAssetAccount,
    "FreezeAssetFrozen": transaction_field.FreezeAssetFrozen,
    "GlobalNumUint": transaction_field.GlobalNumUint,
    "GlobalNumByteSlice": transaction_field.GlobalNumByteSlice,
    "LocalNumUint": transaction_field.LocalNumUint,
    "LocalNumByteSlice": transaction_field.LocalNumByteSlice,
    "ExtraProgramPages": transaction_field.ExtraProgramPages,
    "Nonparticipation": transaction_field.Nonparticipation,
    "NumLogs": transaction_field.NumLogs,
    "CreatedAssetID": transaction_field.CreatedAssetID,
    "CreatedApplicationID": transaction_field.CreatedApplicationID,
    "LastLog": transaction_field.LastLog,
    "StateProofPK": transaction_field.StateProofPK,
    "NumApprovalProgramPages": transaction_field.NumApprovalProgramPages,
    "NumClearStateProgramPages": transaction_field.NumClearStateProgramPages,
}


ARRAY_TX_FIELD_TO_OBJECT = {
    "Accounts": transaction_field.Accounts,
    "ApplicationArgs": transaction_field.ApplicationArgs,
    "Applications": transaction_field.Applications,
    "Assets": transaction_field.Assets,
    "Logs": transaction_field.Logs,
    "ApprovalProgramPages": transaction_field.ApprovalProgramPages,
    "ClearStateProgramPages": transaction_field.ClearStateProgramPages,
}


def _parse_int(x: str) -> int:
    """Parse teal integers.

    Teal supports three formats to write integers, hex, octal and
    decimal. hexadecimal numbers start with the prefix 0x and octal
    numbers have prefix 0.

    Args:
        x: string representation of the teal integer.

    Returns:
        python integer equal to the value represented by the given
        teal integer.
    """

    if x.startswith("0x"):
        return int(x[2:], 16)
    if x.startswith("0"):
        return int(x, 8)
    return int(x)


def parse_transaction_field(tx_field: str, use_stack: bool) -> transaction_field.TransactionField:
    """Parse transaction fields.

    Args:
        tx_field: string representation of the field.
        use_stack: boolean representing whether the array transaction field
            takes it's index from stack instead of as immediate argument.

    Returns:
        object of class corresponding to the given transaction field.
    """
    # parse array transaction fields
    for field, obj in ARRAY_TX_FIELD_TO_OBJECT.items():
        if tx_field.startswith(field):
            index = -1 if use_stack else _parse_int(tx_field[len(field) + 1 :])  # +1 for space(" ")
            return obj(index)

    tx_field = tx_field.replace(" ", "")
    return TX_FIELD_TXT_TO_OBJECT[tx_field]()



"""Printer for exporting subroutines in dot format.

Control Flow Graph of each subroutine defined in the contract
is exported in dot format. As the subroutines are only
supported from teal version 4, for contracts written in version
3 or less, printer doesn't do anything. Each subroutine is saved
in a new dot file with filename ``function_{subroutine_name}_cfg.dot``
in the destination directly which will be current directory if not
specified.

Classes:
    PrinterFunctionCFG: Printer to export CFGs of subroutines defined
        in the contract. The CFGs will be saved in dot format.
"""

import os
from pathlib import Path

from tealer.printers.abstract_printer import AbstractPrinter
from tealer.utils.output import all_subroutines_to_dot, ROOT_OUTPUT_DIRECTORY


class PrinterFunctionCFG(AbstractPrinter):  # pylint: disable=too-few-public-methods
    """Printer to export CFGs of subroutines in dot.

    This printer is supposed to work for contracts written in teal version 4 or greater.
    Dot files will be saved as `function_{subroutine_name}_cfg.dot` in destination folder
    if given or else in the current directory.

    """

    NAME = "subroutine-cfg"
    HELP = "Export the CFG of each subroutine"
    WIKI_URL = "https://github.com/crytic/tealer/wiki/Printer-documentation#subroutine-cfg"

    def print(self) -> None:
        """Export CFG of each subroutine defined in the contract."""
        # outputs multiple files: set the dir to {ROOT_DIRECTORY}/{CONTRACT_NAME}/{"print-"PRINTER_NAME}
        dest = ROOT_OUTPUT_DIRECTORY / Path(self.teal.contract_name) / Path(f"print-{self.NAME}")
        os.makedirs(dest, exist_ok=True)

        all_subroutines_to_dot(self.teal, dest)



from typing import List

from tealer.teal.instructions import instructions, transaction_field
from tealer.detectors.all_detectors import CanCloseAsset
from tealer.teal import global_field

from tests.utils import construct_cfg


CAN_CLOSE_ASSET = """
#pragma version 2
txn Receiver
addr 6ZIOGDXGSQSL4YINHLKCHYRV64FSN4LTUIQ6A4VWYK36FXFF42VI2UV7SM
==
bz wrongreceiver 
txn Fee
int 10000
<
bz highfee
txn RekeyTo
global ZeroAddress
==
bz rekeying
txn CloseRemainderTo
global ZeroAddress
==
bz closing_account
global GroupSize
int 1
==
bz unexpected_group_size
int 1
return
wrongreceiver:
highfee:
rekeying:
closing_account:
unexpected_group_size:
    err
"""

ins_list = [
    instructions.Pragma(2),
    instructions.Txn(transaction_field.Receiver()),
    instructions.Addr("6ZIOGDXGSQSL4YINHLKCHYRV64FSN4LTUIQ6A4VWYK36FXFF42VI2UV7SM"),
    instructions.Eq(),
    instructions.BZ("wrongreceiver"),
    instructions.Txn(transaction_field.Fee()),
    instructions.Int(10000),
    instructions.Less(),
    instructions.BZ("highfee"),
    instructions.Txn(transaction_field.RekeyTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("rekeying"),
    instructions.Txn(transaction_field.CloseRemainderTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("closing_account"),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(1),
    instructions.Eq(),
    instructions.BZ("unexpected_group_size"),
    instructions.Int(1),
    instructions.Return(),
    instructions.Label("wrongreceiver"),
    instructions.Label("highfee"),
    instructions.Label("rekeying"),
    instructions.Label("closing_account"),
    instructions.Label("unexpected_group_size"),
    instructions.Err(),
]

ins_partitions = [
    (0, 5),
    (5, 9),
    (9, 13),
    (13, 17),
    (17, 21),
    (21, 23),
    (23, 24),
    (24, 25),
    (25, 26),
    (26, 27),
    (27, 29),
]
bbs_links = [
    (0, 1),
    (0, 6),
    (1, 2),
    (1, 7),
    (2, 3),
    (2, 8),
    (3, 4),
    (3, 9),
    (4, 5),
    (4, 10),
    (6, 7),
    (7, 8),
    (8, 9),
    (9, 10),
]

bbs = construct_cfg(ins_list, ins_partitions, bbs_links)

CAN_CLOSE_ASSET_VULNERABLE_PATHS = [
    [bbs[0], bbs[1], bbs[2], bbs[3], bbs[4], bbs[5]],
]


CAN_CLOSE_ASSET_LOOP = """
#pragma version 4
txn Receiver
addr 6ZIOGDXGSQSL4YINHLKCHYRV64FSN4LTUIQ6A4VWYK36FXFF42VI2UV7SM
==
bz wrongreceiver 
txn Fee
int 10000
<
bz highfee
txn RekeyTo
global ZeroAddress
==
bz rekeying
txn CloseRemainderTo
global ZeroAddress
==
bz closing_account
global GroupSize
int 1
==
bz unexpected_group_size
int 0
b loop
wrongreceiver:
highfee:
rekeying:
closing_account:
unexpected_group_size:
    err
loop:
    dup
    int 5
    <
    bz end
    int 1
    +
    b loop
end:
    int 1
    return
"""

ins_list = [
    instructions.Pragma(4),
    instructions.Txn(transaction_field.Receiver()),
    instructions.Addr("6ZIOGDXGSQSL4YINHLKCHYRV64FSN4LTUIQ6A4VWYK36FXFF42VI2UV7SM"),
    instructions.Eq(),
    instructions.BZ("wrongreceiver"),
    instructions.Txn(transaction_field.Fee()),
    instructions.Int(10000),
    instructions.Less(),
    instructions.BZ("highfee"),
    instructions.Txn(transaction_field.RekeyTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("rekeying"),
    instructions.Txn(transaction_field.CloseRemainderTo()),
    instructions.Global(global_field.ZeroAddress()),
    instructions.Eq(),
    instructions.BZ("closing_account"),
    instructions.Global(global_field.GroupSize()),
    instructions.Int(1),
    instructions.Eq(),
    instructions.BZ("unexpected_group_size"),
    instructions.Int(0),
    instructions.B("loop"),
    instructions.Label("wrongreceiver"),
    instructions.Label("highfee"),
    instructions.Label("rekeying"),
    instructions.Label("closing_account"),
    instructions.Label("unexpected_group_size"),
    instructions.Err(),
    instructions.Label("loop"),
    instructions.Dup(),
    instructions.Int(5),
    instructions.Less(),
    instructions.BZ("end"),
    instructions.Int(1),
    instructions.Add(),
    instructions.B("loop"),
    instructions.Label("end"),
    instructions.Int(1),
    instructions.Return(),
]

ins_partitions = [
    (0, 5),
    (5, 9),
    (9, 13),
    (13, 17),
    (17, 21),
    (21, 23),
    (23, 24),
    (24, 25),
    (25, 26),
    (26, 27),
    (27, 29),
    (29, 34),
    (34, 37),
    (37, 40),
]

bbs_links = [
    (0, 1),
    (0, 6),
    (1, 2),
    (1, 7),
    (2, 3),
    (2, 8),
    (3, 4),
    (3, 9),
    (4, 5),
    (4, 10),
    (6, 7),
    (7, 8),
    (8, 9),
    (9, 10),
    (5, 11),
    (11, 12),
    (11, 13),
    (12, 11),
]

bbs = construct_cfg(ins_list, ins_partitions, bbs_links)

CAN_CLOSE_ASSET_LOOP_VULNERABLE_PATHS = [
    [bbs[0], bbs[1], bbs[2], bbs[3], bbs[4], bbs[5], bbs[11], bbs[13]]
]


can_close_asset_tests = [
    (CAN_CLOSE_ASSET, CanCloseAsset, CAN_CLOSE_ASSET_VULNERABLE_PATHS),
    (CAN_CLOSE_ASSET_LOOP, CanCloseAsset, CAN_CLOSE_ASSET_LOOP_VULNERABLE_PATHS),
]


CAN_CLOSE_ASSET_GROUP_INDEX_0 = """
#pragma version 6
txn GroupIndex
int 0
!=
bnz index_not_0
gtxn 0 AssetCloseTo  // index must be 0
global ZeroAddress
==
assert
b process_txn
index_not_0:
global GroupSize // index is not zero and groupsize is 2. So, txn index must be 1
int 2
==
assert
gtxn 1 AssetCloseTo
global ZeroAddress
==
assert
process_txn:
int 1
return
"""

CAN_CLOSE_ASSET_GROUP_INDEX_0_VULNERABLE_PATHS: List[List[int]] = []  # not vulnerable


CAN_CLOSE_ASSET_GROUP_INDEX_1 = """
#pragma version 6
txn GroupIndex
int 0
!=
bnz index_not_0
gtxn 0 AssetCloseTo  // index must be 0
addr AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEVAL4QAJS7JHB4
==
assert
b process_txn
index_not_0:
global GroupSize    // index is not zero and groupsize is **3**. So, txn index can be 1 or 2
int 3
==
assert
gtxn 1 AssetCloseTo     // only checks gtxn 1. so, is vulnerable if txn index is 2
global ZeroAddress
==
assert
process_txn:
int 1
return
"""

CAN_CLOSE_ASSET_GROUP_INDEX_1_VULNERABLE_PATHS: List[List[int]] = [[0, 2, 3]]


CAN_CLOSE_ASSET_GROUP_INDEX_2 = """
#pragma version 6
global GroupSize
int 2
!=
bnz group_size_is_not_2
gtxn 0 AssetCloseTo  // group size is 2. so check index 0 and 1.
addr SKSOWAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEVAL4QAJS7JHB4
==
assert
gtxn 1 AssetCloseTo
global ZeroAddress
==
assert
b process_txn
group_size_is_not_2:
global GroupSize    // group_size is not 2
int 3
<
assert              // group_size is less than 3 and is not 2. So, it must be 1 => index must be 0
gtxn 0 AssetCloseTo    
global ZeroAddress
!=
bnz fail_txn
process_txn:
int 1
return
fail_txn:
err
"""

CAN_CLOSE_ASSET_GROUP_INDEX_2_VULNERABLE_PATHS: List[List[int]] = []  # not vulnerable


CAN_CLOSE_ASSET_GROUP_INDEX_3 = """
#pragma version 6
global GroupSize
int 2
!=
bnz group_size_is_not_2
gtxn 0 AssetCloseTo  // group size is 2. so check index 0 and 1.
global ZeroAddress
==
assert
gtxn 1 AssetCloseTo
global ZeroAddress
==
assert
b process_txn
group_size_is_not_2:
global GroupSize    // group_size is not 2
int 3
==
assert              // group_size is 3. check at indices 0, 1, 2
gtxn 0 AssetCloseTo    
global ZeroAddress
==
bz fail_txn
gtxn 2 AssetCloseTo
global ZeroAddress
==
bz fail_txn
gtxn 1 AssetCloseTo
global ZeroAddress
==
bz fail_txn
process_txn:
int 1
return
fail_txn:
err
"""

CAN_CLOSE_ASSET_GROUP_INDEX_3_VULNERABLE_PATHS: List[List[int]] = []  # not vulnerable


new_can_close_asset_tests = [
    (CAN_CLOSE_ASSET_GROUP_INDEX_0, CanCloseAsset, CAN_CLOSE_ASSET_GROUP_INDEX_0_VULNERABLE_PATHS),
    (CAN_CLOSE_ASSET_GROUP_INDEX_1, CanCloseAsset, CAN_CLOSE_ASSET_GROUP_INDEX_1_VULNERABLE_PATHS),
    (CAN_CLOSE_ASSET_GROUP_INDEX_2, CanCloseAsset, CAN_CLOSE_ASSET_GROUP_INDEX_2_VULNERABLE_PATHS),
    (CAN_CLOSE_ASSET_GROUP_INDEX_3, CanCloseAsset, CAN_CLOSE_ASSET_GROUP_INDEX_3_VULNERABLE_PATHS),
]



"""Printer for exporting full contract CFG in dot format.

Classes:
    PrinterCFG: Printer to export contract CFG.
"""

import os
from pathlib import Path

from tealer.printers.abstract_printer import AbstractPrinter
from tealer.utils.output import full_cfg_to_dot, ROOT_OUTPUT_DIRECTORY


class PrinterCFG(AbstractPrinter):  # pylint: disable=too-few-public-methods
    """Printer to export contract CFG in dot."""

    NAME = "cfg"
    HELP = "Export the CFG of entire contract"
    WIKI_URL = "https://github.com/crytic/tealer/wiki/Printer-documentation#cfg"

    def print(self) -> None:
        """Export the CFG of entire contract."""
        # outputs a single file: set the dir to {ROOT_DIRECTORY}/{CONTRACT_NAME}
        dest = ROOT_OUTPUT_DIRECTORY / Path(self.teal.contract_name)
        os.makedirs(dest, exist_ok=True)

        filename = dest / Path("full_cfg.dot")

        print(f"\nCFG exported to file: {filename}")
        full_cfg_to_dot(self.teal, filename=filename)



from typing import Tuple, List, Type
from pathlib import Path
import pytest


from tealer.detectors.abstract_detector import AbstractDetector
from tealer.detectors.all_detectors import (
    MissingRekeyTo,
    MissingFeeCheck,
    CanCloseAccount,
    CanCloseAsset,
    IsUpdatable,
    IsDeletable,
    AnyoneCanDelete,
    AnyoneCanUpdate,
)
from tealer.utils.command_line.group_config import read_config_from_file
from tealer.utils.command_line.common import init_tealer_from_config
from tealer.utils.output import GroupTransactionOutput

from tests.group_transactions.utils import expected_output_from_file


# config, detector, expected output
ALL_TESTS: List[Tuple[Path, Type[AbstractDetector], Path]] = [
    (
        Path("tests/group_transactions/basic/config.yaml"),
        MissingRekeyTo,
        Path("tests/group_transactions/basic/expected_output.yaml"),
    ),
    (
        Path("tests/group_transactions/basic/config.yaml"),
        MissingFeeCheck,
        Path("tests/group_transactions/basic/expected_output.yaml"),
    ),
    (
        Path("tests/group_transactions/basic/config.yaml"),
        CanCloseAccount,
        Path("tests/group_transactions/basic/expected_output_close_to.yaml"),
    ),
    (
        Path("tests/group_transactions/basic/config.yaml"),
        CanCloseAsset,
        Path("tests/group_transactions/basic/expected_output_asset_close_to.yaml"),
    ),
    (
        Path("tests/group_transactions/basic/config.yaml"),
        IsUpdatable,
        Path("tests/group_transactions/basic/expected_output_update_delete.yaml"),
    ),
    (
        Path("tests/group_transactions/basic/config.yaml"),
        IsDeletable,
        Path("tests/group_transactions/basic/expected_output_update_delete.yaml"),
    ),
    (
        Path("tests/group_transactions/basic/config.yaml"),
        AnyoneCanUpdate,
        Path("tests/group_transactions/basic/expected_output_update_delete.yaml"),
    ),
    (
        Path("tests/group_transactions/basic/config.yaml"),
        AnyoneCanDelete,
        Path("tests/group_transactions/basic/expected_output_update_delete.yaml"),
    ),
]


# pylint: disable=too-many-locals
@pytest.mark.parametrize("test", ALL_TESTS)  # type: ignore
def test_instructions_output_detectors(test: Tuple[Path, Type[AbstractDetector], Path]) -> None:
    config_path, detector, expected_output_file_path = test

    expected_output = expected_output_from_file(expected_output_file_path)
    tealer = init_tealer_from_config(read_config_from_file(config_path))
    tealer.register_detector(detector)
    output = tealer.run_detectors()[0]

    reported_output: List["GroupTransactionOutput"] = []
    for i in output:
        assert isinstance(i, GroupTransactionOutput)
        reported_output.append(i)
        i.generate_output(Path("."))

    # expected output is also sorted by operation name
    reported_output = sorted(reported_output, key=lambda x: x.group_transaction.operation_name)
    expected_output_operations = sorted(expected_output.operations, key=lambda x: x.operation)
    # same number of groups/operations are reported as vulnerable
    assert len(reported_output) == len(expected_output_operations)

    for expected_operation, reported_operation in zip(expected_output_operations, reported_output):
        # reported same operation
        assert expected_operation.operation == reported_operation.group_transaction.operation_name
        # reported same transactions
        assert len(expected_operation.vulnerable_transactions) == len(
            reported_operation.transactions.keys()
        )
        expected_transactions = sorted(
            expected_operation.vulnerable_transactions, key=lambda x: x.transaction_id
        )
        reported_transactions = sorted(
            list(reported_operation.transactions.keys()), key=lambda x: x.transacton_id
        )

        for expected_txn, reported_txn in zip(expected_transactions, reported_transactions):
            # same transaction id
            assert expected_txn.transaction_id == reported_txn.transacton_id
            # same contracts
            expected_contracts = sorted(
                expected_txn.contracts, key=lambda x: (x.contract_name, x.function_name)
            )
            reported_contracts = sorted(
                reported_operation.transactions[reported_txn],
                key=lambda x: (x.contract.contract_name, x.function_name),
            )

            assert len(expected_contracts) == len(reported_contracts)
            for expected_contract, reported_contract in zip(expected_contracts, reported_contracts):
                # same contract name and function name
                assert expected_contract.contract_name == reported_contract.contract.contract_name
                assert expected_contract.function_name == reported_contract.function_name

    # for i in output:
    #     # assert isinstance(output, Gr)
    #     out = i.to_json()
    #     print(out["operation"])
    #     print(out["transactions"])
    # assert(False)



"""
The main purpose of transaction-context analysis is to simplify detectors based on transaction field validations.

Transaction field based detectors:
    Detectors that report execution paths that do not validate the value of a problematic transaction field such as
    RekeyTo, CloseRemainderTo, OnCompletion, etc...

The transaction field of the transaction executing the target contract can be accessed in multiple ways:

- The target contract uses `txn {field}`.
- The target contract use its index in the group and access the transaction field using gtxn.. type instructions.
- A different contract uses the target contract's group index and access the target transaction field using gtxn.. type instructions.

Code patterns for accessing the field:

Index types:

    - Absolute index: The index of the target transaction is given by the user input or is fixed at compile time.
    - Relative index: The user input or the compiled code contains the offset. The index is calculated using `Txn.group_index() +/- offset`.

The code patterns for target contract validating its own transaction fields:

- The field is accessed using `txn {field}`
- The target contract validates its index and uses absolute index:
    - The absolute index is fixed at compile time and `gtxn {index} {field}` is used.
    - The absolute index is fixed at compile time but is passed from one location to other:
        - The absolute index is pushed to stack using an `int` instruction and `gtxns {field}` is used.
            - The int instruction is pushed in the same basic block.
            - The int instruction is pushed in different basic block but the basic block is of the
                same subroutine.
            - The int instruction value is passed as a subroutine's argument.
    - The index is computed using `Txn.group_index()` and `gtxns {field}` is used.
    - The index is an user input.
- The target contract uses the group size (`global GroupSize`) and iterates over all transactions in the group
    - accesses all indexes in the group sequentially, the index is fixed at compile time.
        - Above listed code patterns for compile time fixed indexes.
    - Iterates in a loop
    - index is supplied as a subroutine argument.

The code patterns for other contracts validating the target's transaction fields:

- The other contract uses the group size (`global GroupSize`) and iterates over all transactions in the group
    - Above listed code patterns
- The other contracts validates the target contract's index and uses absolute index.
    - The absolute index is fixed at compile time and `gtxn {index} {field}` is used.
    - The absolute index is fixed at compile time but is passed from one location to other:
        - Above listed patterns
- The other contract uses relative index and accesses target's contract fields.
    - The offset is fixed at compile time
        - The index is calculated immediately after `int` instruction pushes the offset: relative index = Txn.group_index +/- offset.
            - The calculated index is used by `gtxns {field}` instruction in the same block.
            - The calculated index is passed around to different blocks.
            - The calculated index is passed as argument to a subroutine.
        - The offset is passed around to different blocks, to a subroutine.
            - Above listed patterns.
    - The offset is user input
        - Above listed patterns.

Code patterns for comparing the field:

- The accessed field is directly compared against the target value.
- The accessed field is compared against a value which reduces the list of possible values.
- The value of the accessed field is passed around:
    - basic blocks of the same subroutine.
    - is passed to an subroutine which compares the field.
- The value is compared using transitive properties: e.g `Gtxn[0].fee + Gtxn[1].fee < 3000`, fees of both transactions are compared.

Code patterns for asserting the comparison:

- The result of the comparison is directly asserted using `assert`.
- The result of the comparison is used by conditional branching instructions `bz`, `bnz`.
- The results of multiple comparisons are combined using `&&`, `||` and `!`.
    - The combined result is asserted using `assert`
    - The combined result is used by `bz` or `bnz` instructions.


The contract analyzed by Tealer can use combination of any of the above patterns. The transaction field analysis tries to
cover as many combinations as possible.

The goal of the transaction field analysis:

- Given an contract, it gives the possible values for the target transaction field of the target contract and possible values
    for the target transaction field of other contracts in the group.

The current results of transaction field analysis:

- Given an contract, it gives the possible values for the target transaction field of the target contract.
    Only some of the patterns are supported.


How is it implemented:

DataflowTransactionContext (in generic.py)

The generic.py contains the utilities that are needed for any transaction field analysis.

The execution flow for an analysis is as follows:

- DataflowTransactionContext.run_analysis()
    - Compute possible values for the transaction field at basic block level (DataflowTransactionContext._block_level_constraints()).
        - i.e if the contract would contain just this basic block, then what would be the possible
            transaction fields.
    - Compute possible values for the transaction field at path level (DataflowTransactionContext._path_level_constraints())
        - This is used when comparison result is used by `bz`, `bnz` instructions.
    - Combine the transaction field information of all basic blocks and edges in the CFG.
        - This is done using iterative fixed-point algorithm.
    - The end result:
        - For each block, list of possible values for the transaction field such that
            if and only if the transaction field during execution is one of the possible values:
            - The execution might reach this basic block
            - The execution might reach end of the basic block
            - The execution might reach from this basic block to a leaf block.

DataflowTransactionContext._block_level_constraints() computes possible values at block level.
    - finds `assert`, `return` instructions in the block.
    - Uses DataflowTransactionContext._get_asserted() on the input argument of `assert` and `return` instructions
        to find possible values for the field.

DataflowTransactionContext._path_level_constraints() computes possible values at path level.
    - checks the basic block has conditional branch instruction: `bz` or `bnz`.
    - Uses DataflowTransactionContext._get_asserted() on the input argument of `bz` or `bnz` to find
        possible values for the field for each out-edge(next) of the block.

DataflowTransactionContext._get_asserted() computes possible values for the field given an comparison result.
    - The input expression might contain `&&`, `||` or `!` and the result of the expression depends on multiple
        small comparisons.
    - This function finds all individual comparison expressions and uses DataflowTransactionContext._get_asserted_single()
        on each individual comparisons.
    - Given the possible values for each individual comparison, this function combines them to find possible values
        for the input comparison.

DataflowTransactionContext._get_asserted_single() is specific for individual fields and is implemented by the corresponding classes.

What are possible valus returned by _get_asserted() and _get_asserted_single()?

The returned values is a Tuple containing `true_values` and `false_values`.

true_values: Given that the execution of input expression results in non-zero value on top of stack, What values
    can the transaction field have?
false_values: Given that the execution of input expression results in zero value on top of stack, What values
    can the transaction field have?

For example input expression is `Txn.group_index() == 2`, then
    - if the result is non-zero => group_index must be in [2] => true_values for group_index = [2]
    - if the result is zero => group_index must be in [0, 1, 3, ..., 15] => false_values for group_index = [0, 1, 3, ..., 15]

_get_asserted_single() returns true_values and false_values for a single direct comparison.
_get_asserted() combines true_values and false_values for all individual direct comparisons and returns true_values
    and false_values for the entire input_expression.

How are these true_values and false_values used by _block_level_constraints() and _path_level_constraints?

_block_level_constraints() finds following instructions:
    - `assert`: Finds the true_values for input expression of `assert` and sets them as possible values
        For the execution to pass => the result must be non-zero => the transaction field can only have one of true_values to not fail the execution.
    - `return`: Finds the true_values for input expression of `return` and sets them as possible values
        For the execution to succeed => the return value must be non-zero => the expression result must be non-zero => the transaction field can have
        only one of true_values.

_path_level_constraints()
    - finds conditional branch instructions (`bz` or `bnz`)
    - finds the edge taken when the result of input expression is non-zero and the edge taken when the result of expression is zero.
    - Assign true_values as possible values for edge taken when the result of input expression is non-zero
    - Assign false_values as possible values for edge taken when the result of input expression is zero

_block_level_constraints fills the possible values for each block.
_path_level_constraints fills the possible values for each edge.

DataflowTransactionContext.run_analysis uses iterative fixed point algorithm to combine information at each block and edge using the information
for predecessors and successors. This gives the above mentioned end_result.


DataflowTransactionContext.BASE_KEYS and DataflowTransactionContext.KEYS_WITH_GTXN

BASE_KEYS: A key represents a transaction field. The possible values for a transaction field are tracked using its corresponding base key.
    The main reason is: The calculation of true_values and false_values for multiple keys can be same. Using multiple keys
    allows us to not create multiple classes that does exactly the same. The _get_asserted_single() is given the "key" and it
    can compute the values for the transaction field corresponding to the key.

    For example, all address fields have similar implementation for computing the true_values and false_values. A single implementation can
    can be used for all fields.

KEYS_WITH_GTXN: KEYS_WITH_GTXN is a subset of BASE_KEYS.
    For a given field multiple types of possible values can be calculated.
    The possible values of the field of the transaction
        1. executing the contract/function irrespective of the transaction's group index.
        2. executing the contract when the transaction group index is a particular value.
        3. at a given absolute index irrespective of which contract is executed.
        4. at a given relative index irrespective of which contract is executed.

    KEYS_WITH_GTXN contains the BASE_KEYS for which the 2., 3. and 4. type of values should also be computed.
    For example, computing the 2., 3. and 4. type of information does not make sense for GroupSize and GroupIndex
    fields. These fields will not be included in KEYS_WITH_GTXN

ANALYSIS_KEYS: List of all keys used for the analysis.
ANALYSIS_KEYS contains BASE_KEYS and keys to represent type 2., 3. and 4. values for each key in KEYS_WITH_GTXN.

How are values for each type of key calculated?

For type (1), The comparisons made on the value accessed using `txn {field}` are used.

For type (2), ANALYSIS_KEYS contains a key for each KEYS_WITH_GTXN and all possible indices.
    - The analysis key with index `i` and `field` key of KEYS_WITH_GTXN will store information of type (2)
    - First, the information at block level and path level is computed using the comparisons made on the field
        accessed using an `gtxn` instruction with absolute index `i`.
    - The information will be of type (3) at this step. In order to make it of type (2), the type (1) information
        is combined with the information at block level and path level.
    - see DataflowTransactionContext._update_gtxn_constraints()


For every field type (1) is computed before type (2).
Computation of type (2) information requires the possible group indices, so possible group indices are computed before
any other field is analyzed.

Supported fields

addr_fields.AddrFields handles support of all address type transaction fields. Add the field name to BASE_KEYS to collect values
for a field.

fee_field.FeeField supports Fee field.

int_fields.GroupIndices supports GroupIndex and GroupSize fields.

txn_types.TxnType supports Transaction type.
"""



"""Collects and exports printers defined in tealer.

Exported printers are:

* ``human-summary``: PrinterHumanSummary prints summary of the contract.
* ``call-graph``: PrinterCallGraph exports call-graph of the contract.
* ``function-cfg``: PrinterFunctionCFG exports contract subroutines in dot format.
* ``cfg``: PrinterCFG exports CFG of the entire contract.
"""

# pylint: disable=unused-import
from tealer.printers.human_summary import PrinterHumanSummary
from tealer.printers.call_graph import PrinterCallGraph
from tealer.printers.function_cfg import PrinterFunctionCFG
from tealer.printers.full_cfg import PrinterCFG
from tealer.printers.transaction_context import PrinterTransactionContext



# pylint: disable=undefined-variable
# type: ignore[name-defined]
from pathlib import Path
from pyteal import *  # pylint: disable=wildcard-import, unused-wildcard-import


def app1_case_1() -> Expr:
    return Return(Int(1))


def app1_case_2() -> Expr:
    return Return(Int(1))


def app1_case_3() -> Expr:
    return Return(Int(1))


def app1_case_4() -> Expr:
    return Return(Int(1))


def app1_case_5() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Txn.rekey_to() == Global.zero_address(),
                    Txn.fee() < Int(10000),
                    Txn.close_remainder_to() == Global.zero_address(),
                    Txn.asset_close_to() == Global.zero_address(),
                    Txn.on_completion() != OnComplete.UpdateApplication,
                    Txn.on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def app1_case_6() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[Txn.group_index()].rekey_to() == Global.zero_address(),
                    Gtxn[Txn.group_index()].fee() < Int(10000),
                    Gtxn[Txn.group_index()].close_remainder_to() == Global.zero_address(),
                    Gtxn[Txn.group_index()].asset_close_to() == Global.zero_address(),
                    Gtxn[Txn.group_index()].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index()].on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def app1_case_7() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[0].rekey_to() == Global.zero_address(),
                    Gtxn[1].rekey_to() == Global.zero_address(),
                    Gtxn[Int(0)].fee() < Int(100000),
                    Gtxn[1].fee() < Int(1000),
                    Gtxn[Int(0)].close_remainder_to() == Global.zero_address(),
                    Gtxn[1].close_remainder_to() == Global.zero_address(),
                    Gtxn[Int(0)].asset_close_to() == Global.zero_address(),
                    Gtxn[1].asset_close_to() == Global.zero_address(),
                    Gtxn[0].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[0].on_completion() != OnComplete.DeleteApplication,
                    Gtxn[Int(1)].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[Int(1)].on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def app1_case_8() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[0].rekey_to() == Global.zero_address(),
                    Gtxn[0].fee() < Int(10000),
                    Gtxn[0].close_remainder_to() == Global.zero_address(),
                    Gtxn[0].asset_close_to() == Global.zero_address(),
                )
            ),
            Assert(
                And(
                    Gtxn[0].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[0].on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def app1_case_9() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[1].rekey_to() == Global.zero_address(),
                    Gtxn[1].fee() < Int(10000),
                    Gtxn[1].close_remainder_to() == Global.zero_address(),
                    Gtxn[1].asset_close_to() == Global.zero_address(),
                )
            ),
            Assert(
                And(
                    Gtxn[1].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[1].on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def app1_case_10() -> Expr:
    return Return(Int(1))


def app1_case_11() -> Expr:
    return Return(Int(1))


def app1_case_12() -> Expr:
    return Return(Int(1))


def app1_case_13() -> Expr:
    return Return(Int(1))


def app1_case_14() -> Expr:
    return Return(Int(1))


def app1_case_15() -> Expr:
    return Return(Int(1))


def app1_case_16() -> Expr:
    return Return(Int(1))


def app1_case_17() -> Expr:
    return Return(Int(1))


def app1_case_18() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[0].rekey_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(4)].rekey_to() == Global.zero_address(),
                    Gtxn[0].fee() < Int(10000),
                    Gtxn[Txn.group_index() + Int(4)].fee() < Int(10000),
                    Gtxn[0].close_remainder_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(4)].close_remainder_to() == Global.zero_address(),
                    Gtxn[0].asset_close_to() == Global.zero_address(),
                    Gtxn[Txn.group_index() + Int(4)].asset_close_to() == Global.zero_address(),
                )
            ),
            Assert(
                And(
                    Gtxn[0].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[0].on_completion() != OnComplete.DeleteApplication,
                    Gtxn[Txn.group_index() + Int(4)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() + Int(4)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def app1_case_19() -> Expr:
    return Seq(
        [
            Assert(Gtxn[Txn.group_index() + Int(4)].rekey_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() + Int(4)].fee() < Int(10000)),
            Assert(Gtxn[Txn.group_index() + Int(4)].asset_close_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() + Int(4)].close_remainder_to() == Global.zero_address()),
            Assert(
                And(
                    Gtxn[Txn.group_index() + Int(4)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() + Int(4)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def app1_case_20() -> Expr:
    return Seq(
        [
            Assert(Gtxn[1].rekey_to() == Global.zero_address()),
            Assert(Gtxn[1].fee() < Int(10000)),
            Assert(Gtxn[1].asset_close_to() == Global.zero_address()),
            Assert(Gtxn[1].close_remainder_to() == Global.zero_address()),
            Assert(
                And(
                    Gtxn[1].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[1].on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def app1_case_21() -> Expr:
    return Seq(
        [
            Assert(Gtxn[Txn.group_index() + Int(2)].rekey_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() + Int(2)].fee() < Int(10000)),
            Assert(Gtxn[Txn.group_index() + Int(2)].asset_close_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() + Int(2)].close_remainder_to() == Global.zero_address()),
            Assert(
                And(
                    Gtxn[Txn.group_index() + Int(2)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() + Int(2)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def app1_case_22() -> Expr:
    return Return(Int(1))


def app1_case_23() -> Expr:
    return Return(Int(1))


def app1_case_24() -> Expr:
    return Return(Int(1))


def app1_case_25() -> Expr:
    return Return(Int(1))


def app1_case_26() -> Expr:
    return Return(Int(1))


def app1_case_27() -> Expr:
    return Return(Int(1))


def app1_case_28() -> Expr:
    return Seq(
        [
            Assert(Gtxn[Txn.group_index() + Int(5)].rekey_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() + Int(5)].fee() < Int(10000)),
            Assert(Gtxn[Txn.group_index() + Int(5)].asset_close_to() == Global.zero_address()),
            Assert(Gtxn[Txn.group_index() + Int(5)].close_remainder_to() == Global.zero_address()),
            Assert(
                And(
                    Gtxn[Txn.group_index() + Int(5)].on_completion()
                    != OnComplete.UpdateApplication,
                    Gtxn[Txn.group_index() + Int(5)].on_completion()
                    != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def app1_case_29() -> Expr:
    return Return(Int(1))


def app1_case_30() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[0].rekey_to() == Global.zero_address(),
                    Gtxn[0].fee() < Int(10000),
                    Gtxn[0].close_remainder_to() == Global.zero_address(),
                    Gtxn[0].asset_close_to() == Global.zero_address(),
                    Gtxn[0].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[0].on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def app1_case_31() -> Expr:
    return Return(Int(1))


def app1_case_32() -> Expr:
    return Return(Int(1))


def app1_case_33() -> Expr:
    return Return(Int(1))


def app1_case_34() -> Expr:
    return Return(Int(1))


def app1_case_35() -> Expr:
    return Return(Int(1))


def app1_case_36() -> Expr:
    return Seq(
        [
            Assert(
                And(
                    Gtxn[4].rekey_to() == Global.zero_address(),
                    Gtxn[4].fee() < Int(10000),
                    Gtxn[4].close_remainder_to() == Global.zero_address(),
                    Gtxn[4].asset_close_to() == Global.zero_address(),
                    Gtxn[4].on_completion() != OnComplete.UpdateApplication,
                    Gtxn[4].on_completion() != OnComplete.DeleteApplication,
                )
            ),
            Return(Int(1)),
        ]
    )


def app1_case_37() -> Expr:
    return Return(Int(1))


def approval_program() -> Expr:
    test_case = Btoi(Txn.application_args[Int(0)])
    return Seq(
        [
            Cond(
                [test_case == Int(1), app1_case_1()],
                [test_case == Int(2), app1_case_2()],
                [test_case == Int(3), app1_case_3()],
                [test_case == Int(4), app1_case_4()],
                [test_case == Int(5), app1_case_5()],
                [test_case == Int(6), app1_case_6()],
                [test_case == Int(7), app1_case_7()],
                [test_case == Int(8), app1_case_8()],
                [test_case == Int(9), app1_case_9()],
                [test_case == Int(10), app1_case_10()],
                [test_case == Int(11), app1_case_11()],
                [test_case == Int(12), app1_case_12()],
                [test_case == Int(13), app1_case_13()],
                [test_case == Int(14), app1_case_14()],
                [test_case == Int(15), app1_case_15()],
                [test_case == Int(16), app1_case_16()],
                [test_case == Int(17), app1_case_17()],
                [test_case == Int(18), app1_case_18()],
                [test_case == Int(19), app1_case_19()],
                [test_case == Int(20), app1_case_20()],
                [test_case == Int(21), app1_case_21()],
                [test_case == Int(22), app1_case_22()],
                [test_case == Int(23), app1_case_23()],
                [test_case == Int(24), app1_case_24()],
                [test_case == Int(25), app1_case_25()],
                [test_case == Int(26), app1_case_26()],
                [test_case == Int(27), app1_case_27()],
                [test_case == Int(28), app1_case_28()],
                [test_case == Int(29), app1_case_29()],
                [test_case == Int(30), app1_case_30()],
                [test_case == Int(31), app1_case_31()],
                [test_case == Int(32), app1_case_32()],
                [test_case == Int(33), app1_case_33()],
                [test_case == Int(34), app1_case_34()],
                [test_case == Int(35), app1_case_35()],
                [test_case == Int(36), app1_case_36()],
                [test_case == Int(37), app1_case_37()],
            ),
        ]
    )


def compile_program(output_file: Path) -> None:
    teal = compileTeal(approval_program(), mode=Mode.Application, version=7)
    with open(output_file, "w", encoding="utf-8") as f:
        f.write(teal)


if __name__ == "__main__":
    compile_program(Path("app_1.teal"))



"""Detector for finding unoptimized usage of constant GTXN."""

from typing import TYPE_CHECKING, List

from tealer.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DetectorType,
)
from tealer.teal.instructions.instructions import Gtxns, Instruction
from tealer.utils.analyses import is_int_push_ins
from tealer.utils.output import InstructionsOutput

if TYPE_CHECKING:
    from tealer.utils.output import ListOutput


class ConstantGtxn(AbstractDetector):  # pylint: disable=too-few-public-methods
    """Detector to find unoptimized usage of constant GTXN"""

    NAME = "constant-gtxn"
    DESCRIPTION = "Unoptimized Gtxn"
    TYPE = DetectorType.STATELESS

    IMPACT = DetectorClassification.OPTIMIZATION
    CONFIDENCE = DetectorClassification.HIGH

    WIKI_URL = "https://github.com/crytic/tealer/wiki/Detector-Documentation#Unoptimized-Gtxn"
    WIKI_TITLE = "Unoptimized Gtxn"
    WIKI_DESCRIPTION = (
        "- Gtxn[Int(1)].field() produces instructions int 1; gtxns field"
        "- Gtxn[1].field() produces only one instruction: gtxn 1 field"
    )
    WIKI_EXPLOIT_SCENARIO = """"""

    WIKI_RECOMMENDATION = """Use Gtxn[idx].field()."""

    def detect(self) -> "ListOutput":
        """Detector to find unoptimized usage of constant GTXN

        Returns:
              The two instructions to optimized
        """

        detector_output: "ListOutput" = []

        for contract in self.tealer.contracts.values():

            all_findings: List[List[Instruction]] = []

            for function in contract.functions.values():
                for block in function.blocks:
                    if len(block.instructions) < 2:
                        continue

                    # Sliding windows
                    for i in range(0, len(block.instructions) - 1):
                        first = block.instructions[i]
                        second = block.instructions[i + 1]

                        # Note: first / second are always assigned, as we check for len(block.instructions) first
                        # pylint: disable=undefined-loop-variable
                        first_ins_is_int, _ = is_int_push_ins(first)
                        # pylint: disable=undefined-loop-variable
                        if first_ins_is_int and isinstance(second, Gtxns):
                            # pylint: disable=undefined-loop-variable
                            all_findings.append([first, second])

            if all_findings:
                detector_output.append(InstructionsOutput(contract, self, all_findings))

        return detector_output



"""Defines abstract base class for detectors in tealer.

This module contains implementation of abstract class which
defines the attributes, common methods and must implement methods
of detectors. All detectors in tealer must inherit from this class
and set attributes, override abstract methods.

Attributes:
    DetectorType: DetectorType is a comparable enumerator. Detector
        type indicates what kind of contracts is that detector supposed
        to work on and find issues in them. DetectorType defines three
        symbols ``STATELESS``, ``STATEFULL``, ``STATEFULLGROUP``.

        ``STATELESS`` detectors work on and find issues in stateless
        contracts(signatures).

        ``STATEFULL`` detectors try to find issues that commonly occur
        in stateful(application) contracts.

        ``STATEFULLGROUP`` detectors work on stateful contracts that are
        supposed to verify security related checks for other transactions/
        contracts in the atomic group. One of the algorand smart contract
        development architecture is to have one contract commonly an
        application that must be included with most of the transaction
        groups interacting with the system. This contract's job is to
        verify the security properties of other transactions in the group
        along with the security properties of it's own transaction.
        ``STATEFULLGROUP`` type of detectors find issues in this kind of
        contracts, whether they are verifying/doing all the related
        security checks for other transactions in the group.

    DETECTOR_TYPE_TXT: Map from DetectorType symbol to it's string representation.

    DetectorClassification: DetectorClassification is a comparable enumerator.
        It defines ``HIGH``, ``MEDIUM``, ``LOW``, ``INFORMATIONAL`` and
        ``OPTIMIZATION``. This class is used to classify detector's impact and
        confidence.

    classification_txt: Map from DetectorClassification symbol to it's string
        representation.

Classes:
    AbstractDetector: Abstract class to represent detectors in tealer.
"""

import abc
from typing import TYPE_CHECKING

from tealer.utils.comparable_enum import ComparableEnum

if TYPE_CHECKING:
    from tealer.tealer import Tealer
    from tealer.utils.output import ListOutput


class IncorrectDetectorInitialization(Exception):
    """Exception class to represent incorrect detector intialization.

    This exception will be used if any of the necessary attributes of
    the AbstractDetector are not set by the inheriting detector class.
    """


# DetectorType is used to specify "Applicability of a detector."
class DetectorType(ComparableEnum):
    # Order by stateful, stateless and stateful, stateless
    STATEFULL = 0
    STATELESS_AND_STATEFULL = 1
    STATELESS = 2
    STATEFULLGROUP = 3

    UNDEFINED = 255


DETECTOR_TYPE_TXT = {
    DetectorType.STATELESS: "Stateless",
    DetectorType.STATEFULL: "Stateful",
    DetectorType.STATELESS_AND_STATEFULL: "Stateless, Stateful",
    DetectorType.STATEFULLGROUP: "StatefulGroup",
}


class DetectorClassification(ComparableEnum):
    HIGH = 0
    MEDIUM = 1
    LOW = 2
    INFORMATIONAL = 3
    OPTIMIZATION = 4

    UNIMPLEMENTED = 999

    def __str__(self) -> str:
        return self.name.title()


classification_txt = {
    DetectorClassification.OPTIMIZATION: "Optimization",
    DetectorClassification.INFORMATIONAL: "Informational",
    DetectorClassification.LOW: "Low",
    DetectorClassification.MEDIUM: "Medium",
    DetectorClassification.HIGH: "High",
}


class AbstractDetector(metaclass=abc.ABCMeta):  # pylint: disable=too-few-public-methods
    """Abstract class to represent detectors in tealer.

    All detectors in tealer must inherit from this class and overrride
    the abstract methods with functionality specific to them. This class
    defines the api methods that must be implemented by every detector.
    ``detect`` method of each detector will be called to execute the
    detector. Every detector must implement ``detect`` method.

    Attributes:
        NAME: Name of the detector. Detector name will be used while
            displaying information about the printer and also for selecting
            the detector from command line.

        DESCRIPTION: Description of the detector about what it tries to find.

        IMPACT: Severity of the issue the detector tries to detect. IMPACT is
            classified into ``HIGH``, ``MEDIUM``, ``LOW``, ``INFORMATIONAL`` and
        ``OPTIMIZATION`` using DetectorClassification enum.

        CONFIDENCE: Precision of the detector, how often the results of this
            detector are true. Confidence is classified into ``HIGH``, ``MEDIUM``
            and ``LOW`` using DetectorClassification enum.

        WIKI_TITLE: Wiki title of the detector/issue.

        WIKI_DESCRIPTION: Wiki description of the detector.

        WIKI_EXPLOIT_SCENARIO: Example exploit scenario exploiting the issue this
            detector tries to find.

        WIKI_RECOMMENDATION: Help or recommendation message on how to avoid, remove
            this issue/vuln from the contract.

    Args:
        teal: Teal instance representing the contract. Detector will analyze the
            contract represented by :teal: for issues/vulnerabilites .

    Raises:
        IncorrectDetectorInitialization: Exception is raised if any of the class
            attributes are not intialized properly by the inheriting detector class.
    """

    NAME = ""  # detector name
    DESCRIPTION = ""
    TYPE = DetectorType.UNDEFINED

    IMPACT: DetectorClassification = DetectorClassification.UNIMPLEMENTED
    CONFIDENCE: DetectorClassification = DetectorClassification.UNIMPLEMENTED

    WIKI_URL = ""
    WIKI_TITLE = ""
    WIKI_DESCRIPTION = ""
    WIKI_EXPLOIT_SCENARIO = ""
    WIKI_RECOMMENDATION = ""

    def __init__(self, tealer: "Tealer"):
        self.tealer = tealer

        if not self.NAME:
            raise IncorrectDetectorInitialization(
                f"NAME is not initialized {self.__class__.__name__}"
            )

        if not self.DESCRIPTION:
            raise IncorrectDetectorInitialization(
                f"DESCRIPTION is not initialized {self.__class__.__name__}"
            )

        if self.TYPE == DetectorType.UNDEFINED:
            raise IncorrectDetectorInitialization(
                f"TYPE is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI_TITLE:
            raise IncorrectDetectorInitialization(
                f"WIKI_TITLE is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI_DESCRIPTION:
            raise IncorrectDetectorInitialization(
                f"WIKI_DESCRIPTION is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI_EXPLOIT_SCENARIO and self.IMPACT not in [
            DetectorClassification.INFORMATIONAL,
            DetectorClassification.OPTIMIZATION,
        ]:
            raise IncorrectDetectorInitialization(
                f"WIKI_EXPLOIT_SCENARIO is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI_RECOMMENDATION:
            raise IncorrectDetectorInitialization(
                f"WIKI_RECOMMENDATION is not initialized {self.__class__.__name__}"
            )

        if not self.WIKI_URL:
            raise IncorrectDetectorInitialization(
                f"WIKI_URL is not initialized {self.__class__.__name__}"
            )

        if self.IMPACT not in [
            DetectorClassification.LOW,
            DetectorClassification.MEDIUM,
            DetectorClassification.HIGH,
            DetectorClassification.INFORMATIONAL,
            DetectorClassification.OPTIMIZATION,
        ]:
            raise IncorrectDetectorInitialization(
                f"IMPACT is not initialized {self.__class__.__name__}"
            )

        if self.CONFIDENCE not in [
            DetectorClassification.LOW,
            DetectorClassification.MEDIUM,
            DetectorClassification.HIGH,
        ]:
            raise IncorrectDetectorInitialization(
                f"CONFIDENCE is not initialized {self.__class__.__name__}"
            )

    @abc.abstractmethod
    def detect(self) -> "ListOutput":
        """Entry method of detector.

        All detectors must override this method with the functionality specific
        to them. This method is the entry point of the detector and will be
        called to execute it.

        Returns:
            Detector results represented in one of the supported types. Currently,
            all detectors work on Control Flow Graph(CFG) of the contract and represent
            the issues in the form of execution path in the CFG. ExecutionPaths class
            is used to represent them.

        # noqa: DAR202
        """



"""Defines classes to represent transaction fields.

Teal supports access to fields of the transaction that invoked the
execution of this contract through multiple instructions.

Each transaction field is represented by a class in tealer. All the
classes representing the fields must inherit from TransactionField class.

"""

# pylint: disable=too-few-public-methods

# https://developer.algorand.org/docs/reference/teal/opcodes/#txn


class TransactionField:
    """Base class representing a transaction field"""

    def __init__(self) -> None:
        self._version: int = 1

    @property
    def version(self) -> int:
        """Teal version this field is introduced in and supported from.

        Returns:
            Teal version the field is introduced in and supported from.
        """
        return self._version

    def __str__(self) -> str:
        return self.__class__.__qualname__


class TransactionArrayField(TransactionField):
    """Base class to represent array transaction fields.

    Array transaction fields have an additional index parameter.
    """

    def __init__(self, idx: int):
        super().__init__()
        self._idx = idx

    @property
    def idx(self) -> int:
        return self._idx

    def __str__(self) -> str:
        s = "" if self._idx < 0 else " " + str(self._idx)
        return self.__class__.__qualname__ + s


class Sender(TransactionField):
    """address of sender of this transaction."""


class Fee(TransactionField):
    """Transaction fee in micro-Algos."""


class FirstValid(TransactionField):
    """minimum round number after which this transaction is valid."""


class LastValid(TransactionField):
    """maximum round number before which this transaction is valid."""


class Note(TransactionField):
    """Any data up to 1024 bytes."""


class Lease(TransactionField):
    """A lease to enforces mutual exclusion of transactions."""


class Receiver(TransactionField):
    """Address of the account that receives the amount."""


class Amount(TransactionField):
    """The total amount to sent in microAlgos."""


class CloseRemainderTo(TransactionField):
    """if set, the account will be closed and all funds will be sent to the given address"""


class VotePK(TransactionField):
    """The root participation public key."""


class SelectionPK(TransactionField):
    """The VRF public key."""


class VoteFirst(TransactionField):
    """The first round the participation key is valid."""


class VoteLast(TransactionField):
    """The last round the participation key is valid."""


class VoteKeyDilution(TransactionField):
    """Dilution for the 2-level participation key."""


class Type(TransactionField):
    """Specifies the type of the transaction"""


class TypeEnum(TransactionField):
    """Type of the transaction"""


class XferAsset(TransactionField):
    """The unique id of the asset to be transferred."""


class AssetAmount(TransactionField):
    """The amount of the asset to be transferred."""


class AssetSender(TransactionField):
    """The sender of the asset transfer."""


class AssetReceiver(TransactionField):
    """The receiver of the asset transfer."""


class AssetCloseTo(TransactionField):
    """if set, removes the asset holding from sender and reduced minimum required balance."""


class GroupIndex(TransactionField):
    """Position of this transaction within the atomic group transaction."""


class TxID(TransactionField):
    """The computed ID for this transaction."""


class ApplicationID(TransactionField):
    """ApplicationID from ApplicationCall transaction"""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class OnCompletion(TransactionField):
    """ApplicationCall transaction on completion action."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class ApplicationArgs(TransactionArrayField):
    """Arguments passed to the application in the ApplicationCall."""

    def __init__(self, idx: int):
        super().__init__(idx)
        self._version: int = 2


class NumAppArgs(TransactionField):
    """Number of Application Args."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class Accounts(TransactionArrayField):
    """Accounts listed in the ApplicationCall transaction."""

    def __init__(self, idx: int):
        super().__init__(idx)
        self._version: int = 2


class NumAccounts(TransactionField):
    """Number of Accounts."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class Applications(TransactionArrayField):
    """Foreign Apps listed in the ApplicationCall transaction."""

    def __init__(self, idx: int):
        super().__init__(idx)
        self._version: int = 3


class NumApplications(TransactionField):
    """Number of Applications."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 3


class Assets(TransactionArrayField):
    """Foreign Assets listed in the ApplicationCall transaction."""

    def __init__(self, idx: int):
        super().__init__(idx)
        self._version: int = 3


class NumAssets(TransactionField):
    """Number of Assets."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 3


class ApprovalProgram(TransactionField):
    """Approval program of the application."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class ClearStateProgram(TransactionField):
    """Clear state program of the application."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class RekeyTo(TransactionField):
    """Sender's new AuthAddr."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class ConfigAsset(TransactionField):
    """AssetID in asset config transaction."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class ConfigAssetTotal(TransactionField):
    """Total number of units of asset created."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class ConfigAssetDecimals(TransactionField):
    """Number of digits to display after the decimal place."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class ConfigAssetDefaultFrozen(TransactionField):
    """Whether the asset's slots are frozen by default or not."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class ConfigAssetUnitName(TransactionField):
    """Unit name of the asset."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class ConfigAssetName(TransactionField):
    """The asset name."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class ConfigAssetURL(TransactionField):
    """Url associated with the asset."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class ConfigAssetMetadataHash(TransactionField):
    """32 byte commitment to some unspecified asset metadata."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class ConfigAssetManager(TransactionField):
    """Manager address, only account that can authorize transactions to re-configure or destroy an asset."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class ConfigAssetReserve(TransactionField):
    """Reserve address, where non-minted assets will reside."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class ConfigAssetFreeze(TransactionField):
    """Freeze account, which is allowed to freeze or unfreeze the asset holding for an account."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class ConfigAssetClawback(TransactionField):
    """Clawback address, which can transfer assets from and to any asset holder."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class FreezeAsset(TransactionField):
    """AssetID being frozen or un-frozen."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class FreezeAssetAccount(TransactionField):
    """address whose asset slot is being frozen or un-frozen."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class FreezeAssetFrozen(TransactionField):
    """The new frozen value, 0 or 1."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 2


class GlobalNumUint(TransactionField):
    """Number of global state integers in ApplicationCall."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 3


class GlobalNumByteSlice(TransactionField):
    """Number of global state byteslices in ApplicationCall."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 3


class LocalNumUint(TransactionField):
    """Number of local state integers in ApplicationCall."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 3


class LocalNumByteSlice(TransactionField):
    """Number of local state byteslices in ApplicationCall."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 3


class ExtraProgramPages(TransactionField):
    """Number of additional pages for each of approval and clear state programs."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 4


class Nonparticipation(TransactionField):
    """Marks an account nonparticipating for rewards."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 5


class Logs(TransactionArrayField):
    """Log messages emitted by an application call(itxn only)."""

    def __init__(self, idx: int):
        super().__init__(idx)
        self._version: int = 5


class NumLogs(TransactionField):
    """Number of log messages(itxn only)."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 5


class CreatedAssetID(TransactionField):
    """Asset ID allocated by the creation of an ASA (itxn only)."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 5


class CreatedApplicationID(TransactionField):
    """ApplicationID allocated by the creation of an application (itxn only)."""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 5


class LastLog(TransactionField):
    """([]byte) The last message emitted. Empty bytes if none were emitted. Application mode only"""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 6


class StateProofPK(TransactionField):
    """([]byte) 64 byte state proof public key"""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 6


class FirstValidTime(TransactionField):
    """(uint64) UNIX timestamp of block before txn.FirstValid. Fails if negative.

    FirstValidTime is present from Teal v1 itself. However, execution would fail if this
    field is used in versions less than v7.
    """

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 7


class NumApprovalProgramPages(TransactionField):
    """(uint64) Number of Approval Program pages"""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 7


class NumClearStateProgramPages(TransactionField):
    """(uint64) Number of clear state program pages"""

    def __init__(self) -> None:
        super().__init__()
        self._version: int = 7


class ApprovalProgramPages(TransactionArrayField):
    """Approval Program as an array of pages"""

    def __init__(self, idx: int):
        super().__init__(idx)
        self._version: int = 7


class ClearStateProgramPages(TransactionArrayField):
    """Approval Program as an array of pages"""

    def __init__(self, idx: int):
        super().__init__(idx)
        self._version: int = 7



"""Detector for finding execution paths missing UpdateApplication."""

from typing import List, TYPE_CHECKING, Tuple

from tealer.detectors.abstract_detector import (
    AbstractDetector,
    DetectorClassification,
    DetectorType,
)
from tealer.detectors.utils import (
    detect_missing_tx_field_validations_group,
    detect_missing_tx_field_validations_group_complete,
)
from tealer.utils.teal_enums import TealerTransactionType
from tealer.utils.output import ExecutionPaths

if TYPE_CHECKING:
    from tealer.teal.basic_blocks import BasicBlock
    from tealer.utils.output import ListOutput
    from tealer.teal.context.block_transaction_context import BlockTransactionContext
    from tealer.teal.teal import Teal


class IsUpdatable(AbstractDetector):  # pylint: disable=too-few-public-methods
    """Detector to find execution paths missing UpdateApplication check.

    Stateful smart contracts(application) can be updated with the new code
    in algorand. If the application transaction of type UpdateApplication is
    approved by the contract then the application's approval, clear programs
    will be replaced by the ones sent along with the transaction. Contracts
    can check the application transaction type using OnCompletion field.

    This detector tries to find execution paths that approve the application
    transactions("return 1") and doesn't check the OnCompletion field against
    UpdateApplication value. Execution paths that only execute if the application
    transaction is not UpdateApplication are excluded.
    """

    NAME = "is-updatable"
    DESCRIPTION = "Upgradable Applications"
    TYPE = DetectorType.STATEFULL

    IMPACT = DetectorClassification.HIGH
    CONFIDENCE = DetectorClassification.HIGH

    WIKI_URL = "https://github.com/crytic/tealer/wiki/Detector-Documentation#upgradable-application"
    WIKI_TITLE = "Upgradable Application"
    WIKI_DESCRIPTION = (
        "Application can be updated by sending an `UpdateApplication` type application call."
    )
    WIKI_EXPLOIT_SCENARIO = """
```py
@router.method(update_application=CallConfig.CALL)
def update_application() -> Expr:
    return Assert(Txn.sender() == Global.creator_address())
```

Creator updates the application and steals all of its assets.
"""

    WIKI_RECOMMENDATION = """
Do not approve `UpdateApplication` type application calls.
"""

    def detect(self) -> "ListOutput":
        """Detect execution paths with missing UpdateApplication check.

        Returns:
            ExecutionPaths instance containing the list of vulnerable execution
            paths along with name, check, impact, confidence and other detector
            information.
        """

        def checks_field(block_ctx: "BlockTransactionContext") -> bool:
            # return False if Txn Type can be UpdateApplication.
            # return True if Txn Type cannot be UpdateApplication.
            return not TealerTransactionType.ApplUpdateApplication in block_ctx.transaction_types

        # there should be a better to decide which function to call ??
        if self.tealer.output_group:
            # mypy complains if the value is returned directly. Uesd the second suggestion mentioned here:
            # https://mypy.readthedocs.io/en/stable/common_issues.html#variance
            return list(
                detect_missing_tx_field_validations_group_complete(self.tealer, self, checks_field)
            )

        # paths_without_check: List[List[BasicBlock]] = detect_missing_tx_field_validations(
        #     self.teal, checks_field
        # )

        # return ExecutionPaths(self.teal, self, paths_without_check)
        output: List[
            Tuple["Teal", List[List["BasicBlock"]]]
        ] = detect_missing_tx_field_validations_group(self.tealer, checks_field)
        detector_output: "ListOutput" = []
        for contract, vulnerable_paths in output:
            detector_output.append(ExecutionPaths(contract, self, vulnerable_paths))

        return detector_output
